import numpy as np
import pytest

import autoarray as aa


class TestCoordinates:
    def test__central_pixel__depends_on_shape_pixel_scale_and_origin(self):

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(0.1, 0.1)
        )
        assert mask.geometry.central_pixel_coordinates == (1, 1)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(4, 4)), pixel_scales=(0.1, 0.1)
        )
        assert mask.geometry.central_pixel_coordinates == (1.5, 1.5)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(5, 3)),
            pixel_scales=(0.1, 0.1),
            origin=(1.0, 2.0),
        )
        assert mask.geometry.central_pixel_coordinates == (2.0, 1.0)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(2.0, 1.0)
        )
        assert mask.geometry.central_pixel_coordinates == (1, 1)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(4, 4)), pixel_scales=(2.0, 1.0)
        )
        assert mask.geometry.central_pixel_coordinates == (1.5, 1.5)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(5, 3)),
            pixel_scales=(2.0, 1.0),
            origin=(1.0, 2.0),
        )
        assert mask.geometry.central_pixel_coordinates == (2, 1)

    def test__centring__adapts_to_max_and_min_of_mask(self):
        mask = np.array(
            [
                [True, True, True, True],
                [True, False, False, True],
                [True, True, True, True],
            ]
        )

        mask = aa.Mask.manual(mask=mask, pixel_scales=(1.0, 1.0))

        assert mask.geometry.mask_centre == (0.0, 0.0)

        mask = np.array(
            [
                [True, True, True, True],
                [True, False, False, False],
                [True, True, True, True],
            ]
        )

        mask = aa.Mask.manual(mask=mask, pixel_scales=(1.0, 1.0))

        assert mask.geometry.mask_centre == (0.0, 0.5)

        mask = np.array(
            [
                [True, True, False, True],
                [True, False, False, True],
                [True, True, True, True],
            ]
        )

        mask = aa.Mask.manual(mask=mask, pixel_scales=(1.0, 1.0))

        assert mask.geometry.mask_centre == (0.5, 0.0)

        mask = np.array(
            [
                [True, True, True, True],
                [False, False, False, True],
                [True, True, True, True],
            ]
        )

        mask = aa.Mask.manual(mask=mask, pixel_scales=(1.0, 1.0))

        assert mask.geometry.mask_centre == (0.0, -0.5)

        mask = np.array(
            [
                [True, True, True, True],
                [True, False, False, True],
                [True, False, True, True],
            ]
        )

        mask = aa.Mask.manual(mask=mask, pixel_scales=(1.0, 1.0))

        assert mask.geometry.mask_centre == (-0.5, 0.0)

        mask = np.array(
            [
                [True, True, True, True],
                [True, False, False, True],
                [False, True, True, True],
            ]
        )

        mask = aa.Mask.manual(mask=mask, pixel_scales=(1.0, 1.0))

        assert mask.geometry.mask_centre == (-0.5, -0.5)

    def test__pixel_grid__y_and_x_ticks(self):
        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(1.0, 1.0)
        )
        assert mask.geometry.yticks == pytest.approx(
            np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(0.5, 0.5)
        )
        assert mask.geometry.yticks == pytest.approx(
            np.array([-0.75, -0.25, 0.25, 0.75]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(6, 3)), pixel_scales=(1.0, 1.0)
        )
        assert mask.geometry.yticks == pytest.approx(
            np.array([-3.0, -1.0, 1.0, 3.0]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 1)), pixel_scales=(1.0, 1.0)
        )
        assert mask.geometry.yticks == pytest.approx(
            np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(1.0, 1.0)
        )
        assert mask.geometry.xticks == pytest.approx(
            np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(0.5, 0.5)
        )
        assert mask.geometry.xticks == pytest.approx(
            np.array([-0.75, -0.25, 0.25, 0.75]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 6)), pixel_scales=(1.0, 1.0)
        )
        assert mask.geometry.xticks == pytest.approx(
            np.array([-3.0, -1.0, 1.0, 3.0]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(1, 3)), pixel_scales=(1.0, 1.0)
        )
        assert mask.geometry.xticks == pytest.approx(
            np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(1.0, 5.0)
        )
        assert mask.geometry.yticks == pytest.approx(
            np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(0.5, 5.0)
        )
        assert mask.geometry.yticks == pytest.approx(
            np.array([-0.75, -0.25, 0.25, 0.75]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(6, 3)), pixel_scales=(1.0, 5.0)
        )
        assert mask.geometry.yticks == pytest.approx(
            np.array([-3.0, -1.0, 1.0, 3.0]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 6)), pixel_scales=(1.0, 5.0)
        )
        assert mask.geometry.yticks == pytest.approx(
            np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(5.0, 1.0)
        )
        assert mask.geometry.xticks == pytest.approx(
            np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(5.0, 0.5)
        )
        assert mask.geometry.xticks == pytest.approx(
            np.array([-0.75, -0.25, 0.25, 0.75]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 6)), pixel_scales=(5.0, 1.0)
        )
        assert mask.geometry.xticks == pytest.approx(
            np.array([-3.0, -1.0, 1.0, 3.0]), 1e-3
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(6, 3)), pixel_scales=(5.0, 1.0)
        )
        assert mask.geometry.xticks == pytest.approx(
            np.array([-1.5, -0.5, 0.5, 1.5]), 1e-3
        )


class TestGrids:
    def test__unmasked_grid__compare_to_array_util(self):

        grid_2d_util = aa.util.grid.grid_2d_via_shape_2d_from(
            shape_2d=(4, 7), pixel_scales=(0.56, 0.56), sub_size=1
        )

        grid_1d_util = aa.util.grid.grid_1d_via_shape_2d_from(
            shape_2d=(4, 7), pixel_scales=(0.56, 0.56), sub_size=1
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(4, 7)), pixel_scales=(0.56, 0.56)
        )
        mask[0, 0] = True

        assert mask.geometry.unmasked_grid.in_1d == pytest.approx(grid_1d_util, 1e-4)
        assert mask.geometry.unmasked_grid.in_2d == pytest.approx(grid_2d_util, 1e-4)
        assert (
            mask.geometry.unmasked_grid.mask == np.full(fill_value=False, shape=(4, 7))
        ).all()

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(1.0, 1.0)
        )

        assert (
            mask.geometry.unmasked_grid.in_2d
            == np.array(
                [
                    [[1.0, -1.0], [1.0, 0.0], [1.0, 1.0]],
                    [[0.0, -1.0], [0.0, 0.0], [0.0, 1.0]],
                    [[-1.0, -1.0], [-1.0, 0.0], [-1.0, 1.0]],
                ]
            )
        ).all()

        grid_2d_util = aa.util.grid.grid_2d_via_shape_2d_from(
            shape_2d=(4, 7), pixel_scales=(0.8, 0.56), sub_size=1
        )

        grid_1d_util = aa.util.grid.grid_1d_via_shape_2d_from(
            shape_2d=(4, 7), pixel_scales=(0.8, 0.56), sub_size=1
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(4, 7)), pixel_scales=(0.8, 0.56)
        )

        assert mask.geometry.unmasked_grid.in_1d == pytest.approx(grid_1d_util, 1e-4)
        assert mask.geometry.unmasked_grid.in_2d == pytest.approx(grid_2d_util, 1e-4)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(1.0, 2.0)
        )

        assert (
            mask.geometry.unmasked_grid.in_2d
            == np.array(
                [
                    [[1.0, -2.0], [1.0, 0.0], [1.0, 2.0]],
                    [[0.0, -2.0], [0.0, 0.0], [0.0, 2.0]],
                    [[-1.0, -2.0], [-1.0, 0.0], [-1.0, 2.0]],
                ]
            )
        ).all()

    def test__grid_with_nonzero_origins__compure_to_array_util(self):

        grid_2d_util = aa.util.grid.grid_2d_via_shape_2d_from(
            shape_2d=(4, 7), pixel_scales=(0.56, 0.56), origin=(1.0, 3.0), sub_size=1
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(4, 7)),
            pixel_scales=(0.56, 0.56),
            origin=(1.0, 3.0),
        )

        assert mask.geometry.unmasked_grid.in_2d == pytest.approx(grid_2d_util, 1e-4)

        grid_1d_util = aa.util.grid.grid_1d_via_shape_2d_from(
            shape_2d=(4, 7), pixel_scales=(0.56, 0.56), origin=(-1.0, -4.0), sub_size=1
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(4, 7)),
            pixel_scales=(0.56, 0.56),
            origin=(-1.0, -4.0),
        )

        assert mask.geometry.unmasked_grid.in_1d == pytest.approx(grid_1d_util, 1e-4)

        grid_2d_util = aa.util.grid.grid_2d_via_shape_2d_from(
            shape_2d=(4, 7), pixel_scales=(0.8, 0.56), origin=(1.0, 2.0), sub_size=1
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(4, 7)),
            pixel_scales=(0.8, 0.56),
            origin=(1.0, 2.0),
        )

        assert mask.geometry.unmasked_grid.in_2d == pytest.approx(grid_2d_util, 1e-4)

        grid_1d_util = aa.util.grid.grid_1d_via_shape_2d_from(
            shape_2d=(4, 7), pixel_scales=(0.8, 0.56), origin=(-1.0, -4.0), sub_size=1
        )

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(4, 7)),
            pixel_scales=(0.8, 0.56),
            origin=(-1.0, -4.0),
        )

        assert mask.geometry.unmasked_grid.in_1d == pytest.approx(grid_1d_util, 1e-4)

    def test__masked_grids_1d(self):

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(1.0, 1.0)
        )

        assert (
            mask.geometry.masked_grid.in_1d
            == np.array(
                [
                    [1.0, -1.0],
                    [1.0, 0.0],
                    [1.0, 1.0],
                    [0.0, -1.0],
                    [0.0, 0.0],
                    [0.0, 1.0],
                    [-1.0, -1.0],
                    [-1.0, 0.0],
                    [-1.0, 1.0],
                ]
            )
        ).all()

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(1.0, 1.0)
        )
        mask[1, 1] = True

        assert (
            mask.geometry.masked_grid.in_1d
            == np.array(
                [
                    [1.0, -1.0],
                    [1.0, 0.0],
                    [1.0, 1.0],
                    [0.0, -1.0],
                    [0.0, 1.0],
                    [-1.0, -1.0],
                    [-1.0, 0.0],
                    [-1.0, 1.0],
                ]
            )
        ).all()

        mask = aa.Mask.manual(
            mask=np.array([[False, True], [True, False], [True, False]]),
            pixel_scales=(1.0, 1.0),
            origin=(3.0, -2.0),
        )

        assert (
            mask.geometry.masked_grid.in_1d
            == np.array([[4.0, -2.5], [3.0, -1.5], [2.0, -1.5]])
        ).all()

    def test__edge_grid(self):
        mask = np.array(
            [
                [True, True, True, True, True, True, True, True, True],
                [True, False, False, False, False, False, False, False, True],
                [True, False, True, True, True, True, True, False, True],
                [True, False, True, False, False, False, True, False, True],
                [True, False, True, False, True, False, True, False, True],
                [True, False, True, False, False, False, True, False, True],
                [True, False, True, True, True, True, True, False, True],
                [True, False, False, False, False, False, False, False, True],
                [True, True, True, True, True, True, True, True, True],
            ]
        )

        mask = aa.Mask.manual(mask=mask, pixel_scales=(1.0, 1.0))

        assert mask.geometry.edge_grid.in_1d[0:11] == pytest.approx(
            np.array(
                [
                    [3.0, -3.0],
                    [3.0, -2.0],
                    [3.0, -1.0],
                    [3.0, -0.0],
                    [3.0, 1.0],
                    [3.0, 2.0],
                    [3.0, 3.0],
                    [2.0, -3.0],
                    [2.0, 3.0],
                    [1.0, -3.0],
                    [1.0, -1.0],
                ]
            ),
            1e-4,
        )

    def test__border_grid(self):
        mask = np.array(
            [
                [True, True, True, True, True, True, True, True, True],
                [True, False, False, False, False, False, False, False, True],
                [True, False, True, True, True, True, True, False, True],
                [True, False, True, False, False, False, True, False, True],
                [True, False, True, False, True, False, True, False, True],
                [True, False, True, False, False, False, True, False, True],
                [True, False, True, True, True, True, True, False, True],
                [True, False, False, False, False, False, False, False, True],
                [True, True, True, True, True, True, True, True, True],
            ]
        )

        mask = aa.Mask.manual(mask=mask, pixel_scales=(1.0, 1.0))

        assert mask.geometry.border_grid.in_1d[0:11] == pytest.approx(
            np.array(
                [
                    [3.0, -3.0],
                    [3.0, -2.0],
                    [3.0, -1.0],
                    [3.0, -0.0],
                    [3.0, 1.0],
                    [3.0, 2.0],
                    [3.0, 3.0],
                    [2.0, -3.0],
                    [2.0, 3.0],
                    [1.0, -3.0],
                    [1.0, 3.0],
                ]
            ),
            1e-4,
        )

    def test__masked_sub_grid(self):
        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)),
            pixel_scales=(1.0, 1.0),
            sub_size=1,
        )

        assert (
            mask.geometry.masked_sub_grid
            == np.array(
                [
                    [1.0, -1.0],
                    [1.0, 0.0],
                    [1.0, 1.0],
                    [0.0, -1.0],
                    [0.0, 0.0],
                    [0.0, 1.0],
                    [-1.0, -1.0],
                    [-1.0, 0.0],
                    [-1.0, 1.0],
                ]
            )
        ).all()

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)),
            pixel_scales=(1.0, 1.0),
            sub_size=2,
        )

        assert (
            mask.geometry.masked_sub_grid
            == np.array(
                [
                    [0.75, -0.75],
                    [0.75, -0.25],
                    [0.25, -0.75],
                    [0.25, -0.25],
                    [0.75, 0.25],
                    [0.75, 0.75],
                    [0.25, 0.25],
                    [0.25, 0.75],
                    [-0.25, -0.75],
                    [-0.25, -0.25],
                    [-0.75, -0.75],
                    [-0.75, -0.25],
                    [-0.25, 0.25],
                    [-0.25, 0.75],
                    [-0.75, 0.25],
                    [-0.75, 0.75],
                ]
            )
        ).all()

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)),
            pixel_scales=(1.0, 1.0),
            sub_size=1,
        )
        mask[1, 1] = True

        assert (
            mask.geometry.masked_sub_grid
            == np.array(
                [
                    [1.0, -1.0],
                    [1.0, 0.0],
                    [1.0, 1.0],
                    [0.0, -1.0],
                    [0.0, 1.0],
                    [-1.0, -1.0],
                    [-1.0, 0.0],
                    [-1.0, 1.0],
                ]
            )
        ).all()

        mask = aa.Mask.manual(
            mask=np.array([[False, True], [True, False], [True, False]]),
            pixel_scales=(1.0, 1.0),
            sub_size=5,
            origin=(3.0, -2.0),
        )

        masked_grid_util = aa.util.grid.grid_1d_via_mask_from(
            mask=mask, pixel_scales=(1.0, 1.0), sub_size=5, origin=(3.0, -2.0)
        )

        assert (mask.geometry.masked_sub_grid == masked_grid_util).all()

    def test__sub_border_1d_grid__compare_numerical_values(self):

        mask = np.array(
            [
                [True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True],
                [True, False, False, True, True, True, True],
                [True, True, True, True, False, True, True],
                [True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True],
            ]
        )

        mask = aa.Mask.manual(mask=mask, pixel_scales=(1.0, 1.0), sub_size=2)

        assert (
            mask.geometry.sub_border_grid_1d
            == np.array([[1.25, -2.25], [1.25, -1.25], [-0.25, 1.25]])
        ).all()

        mask = np.array(
            [
                [True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True],
                [True, True, False, False, False, True, True],
                [True, True, False, False, False, True, True],
                [True, True, False, False, False, True, True],
                [True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True],
            ]
        )

        mask = aa.Mask.manual(mask=mask, pixel_scales=(1.0, 1.0), sub_size=2)

        assert (
            mask.geometry.sub_border_grid_1d
            == np.array(
                [
                    [1.25, -1.25],
                    [1.25, 0.25],
                    [1.25, 1.25],
                    [-0.25, -1.25],
                    [-0.25, 1.25],
                    [-1.25, -1.25],
                    [-1.25, 0.25],
                    [-1.25, 1.25],
                ]
            )
        ).all()


class TestArcsecToPixel:
    def test__pixel_coordinates_from_arcsec_coordinates(self):

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)), pixel_scales=(2.0, 2.0)
        )

        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(1.0, -1.0)
        ) == (0, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(1.0, 1.0)
        ) == (0, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-1.0, -1.0)
        ) == (1, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-1.0, 1.0)
        ) == (1, 1)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(3.0, 3.0)
        )

        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(3.0, -3.0)
        ) == (0, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(3.0, 0.0)
        ) == (0, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(3.0, 3.0)
        ) == (0, 2)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.0, -3.0)
        ) == (1, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.0, 0.0)
        ) == (1, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.0, 3.0)
        ) == (1, 2)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-3.0, -3.0)
        ) == (2, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-3.0, 0.0)
        ) == (2, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-3.0, 3.0)
        ) == (2, 2)

    def test__pixel_coordinates_from_arcsec_coordinates__arcsec_are_pixel_corners(self):

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)), pixel_scales=(2.0, 2.0)
        )

        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(1.99, -1.99)
        ) == (0, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(1.99, -0.01)
        ) == (0, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.01, -1.99)
        ) == (0, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.01, -0.01)
        ) == (0, 0)

        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(2.01, 0.01)
        ) == (0, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(2.01, 1.99)
        ) == (0, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.01, 0.01)
        ) == (0, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.01, 1.99)
        ) == (0, 1)

        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-0.01, -1.99)
        ) == (1, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-0.01, -0.01)
        ) == (1, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-1.99, -1.99)
        ) == (1, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-1.99, -0.01)
        ) == (1, 0)

        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-0.01, 0.01)
        ) == (1, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-0.01, 1.99)
        ) == (1, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-1.99, 0.01)
        ) == (1, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-1.99, 1.99)
        ) == (1, 1)

    def test__pixel_coordinates_from_arcsec_coordinates___arcsec_are_pixel_centres__nonzero_centre(
        self
    ):
        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)),
            pixel_scales=(2.0, 2.0),
            origin=(1.0, 1.0),
        )

        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(2.0, 0.0)
        ) == (0, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(2.0, 2.0)
        ) == (0, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.0, 0.0)
        ) == (1, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.0, 2.0)
        ) == (1, 1)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)),
            pixel_scales=(3.0, 3.0),
            origin=(3.0, 3.0),
        )

        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(6.0, 0.0)
        ) == (0, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(6.0, 3.0)
        ) == (0, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(6.0, 6.0)
        ) == (0, 2)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(3.0, 0.0)
        ) == (1, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(3.0, 3.0)
        ) == (1, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(3.0, 6.0)
        ) == (1, 2)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.0, 0.0)
        ) == (2, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.0, 3.0)
        ) == (2, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.0, 6.0)
        ) == (2, 2)

    def test__pixel_coordinates_from_arcsec_coordinates__arcsec_are_pixel_corners__nonzero_centre(
        self
    ):
        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)),
            pixel_scales=(2.0, 2.0),
            origin=(1.0, 1.0),
        )

        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(2.99, -0.99)
        ) == (0, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(2.99, 0.99)
        ) == (0, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(1.01, -0.99)
        ) == (0, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(1.01, 0.99)
        ) == (0, 0)

        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(3.01, 1.01)
        ) == (0, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(3.01, 2.99)
        ) == (0, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(1.01, 1.01)
        ) == (0, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(1.01, 2.99)
        ) == (0, 1)

        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.99, -0.99)
        ) == (1, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.99, 0.99)
        ) == (1, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-0.99, -0.99)
        ) == (1, 0)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-0.99, 0.99)
        ) == (1, 0)

        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.99, 1.01)
        ) == (1, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(0.99, 2.99)
        ) == (1, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-0.99, 1.01)
        ) == (1, 1)
        assert mask.geometry.pixel_coordinates_from_scaled_coordinates(
            scaled_coordinates=(-0.99, 2.99)
        ) == (1, 1)

    def test__scaled_coordinates_from_pixel_coordinates___scaled_are_pixel_centres__nonzero_centre(
        self
    ):
        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(3.0, 3.0)
        )

        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(0, 0)
        ) == (3.0, -3.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(0, 1)
        ) == (3.0, 0.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(0, 2)
        ) == (3.0, 3.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(1, 0)
        ) == (0.0, -3.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(1, 1)
        ) == (0.0, 0.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(1, 2)
        ) == (0.0, 3.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(2, 0)
        ) == (-3.0, -3.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(2, 1)
        ) == (-3.0, 0.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(2, 2)
        ) == (-3.0, 3.0)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)),
            pixel_scales=(2.0, 2.0),
            origin=(1.0, 1.0),
        )

        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(0, 0)
        ) == (2.0, 0.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(0, 1)
        ) == (2.0, 2.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(1, 0)
        ) == (0.0, 0.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(1, 1)
        ) == (0.0, 2.0)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)),
            pixel_scales=(3.0, 3.0),
            origin=(3.0, 3.0),
        )

        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(0, 0)
        ) == (6.0, 0.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(0, 1)
        ) == (6.0, 3.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(0, 2)
        ) == (6.0, 6.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(1, 0)
        ) == (3.0, 0.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(1, 1)
        ) == (3.0, 3.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(1, 2)
        ) == (3.0, 6.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(2, 0)
        ) == (0.0, 0.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(2, 1)
        ) == (0.0, 3.0)
        assert mask.geometry.scaled_coordinates_from_pixel_coordinates(
            pixel_coordinates=(2, 2)
        ) == (0.0, 6.0)


class TestGridConversions:
    def test__grid_pixels_from_grid_arcsec(self):
        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)), pixel_scales=(2.0, 4.0)
        )

        grid_arcsec_1d = np.array([[1.0, -2.0], [1.0, 2.0], [-1.0, -2.0], [-1.0, 2.0]])

        grid_pixels_util = aa.util.grid.grid_pixels_1d_from(
            grid_scaled_1d=grid_arcsec_1d, shape_2d=(2, 2), pixel_scales=(2.0, 4.0)
        )
        grid_pixels = mask.geometry.grid_pixels_from_grid_scaled_1d(
            grid_scaled_1d=grid_arcsec_1d
        )

        assert (grid_pixels == grid_pixels_util).all()
        assert (grid_pixels.in_1d == grid_pixels_util).all()

    def test__grid_pixel_centres_1d_from_grid_arcsec_1d__same_as_grid_util(self):

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)), pixel_scales=(2.0, 2.0)
        )

        grid_arcsec_1d = np.array([[0.5, -0.5], [0.5, 0.5], [-0.5, -0.5], [-0.5, 0.5]])

        grid_pixels_util = aa.util.grid.grid_pixel_centres_1d_from(
            grid_scaled_1d=grid_arcsec_1d, shape_2d=(2, 2), pixel_scales=(2.0, 2.0)
        )

        grid_pixels = mask.geometry.grid_pixel_centres_from_grid_scaled_1d(
            grid_scaled_1d=grid_arcsec_1d
        )

        assert (grid_pixels == grid_pixels_util).all()

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)), pixel_scales=(7.0, 2.0)
        )

        grid_arcsec_1d = np.array([[1.0, -2.0], [1.0, 2.0], [-1.0, -2.0], [-1.0, 2.0]])

        grid_pixels_util = aa.util.grid.grid_pixel_centres_1d_from(
            grid_scaled_1d=grid_arcsec_1d, shape_2d=(2, 2), pixel_scales=(7.0, 2.0)
        )

        grid_pixels = mask.geometry.grid_pixel_centres_from_grid_scaled_1d(
            grid_scaled_1d=grid_arcsec_1d
        )

        assert (grid_pixels == grid_pixels_util).all()

    def test__grid_pixel_indexes_1d_from_grid_arcsec_1d__same_as_grid_util(self):
        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)), pixel_scales=(2.0, 2.0)
        )

        grid_arcsec = np.array([[1.0, -1.0], [1.0, 1.0], [-1.0, -1.0], [-1.0, 1.0]])

        grid_pixel_indexes_util = aa.util.grid.grid_pixel_indexes_1d_from(
            grid_scaled_1d=grid_arcsec, shape_2d=(2, 2), pixel_scales=(2.0, 2.0)
        )

        grid_pixel_indexes = mask.geometry.grid_pixel_indexes_from_grid_scaled_1d(
            grid_scaled_1d=grid_arcsec
        )

        assert (grid_pixel_indexes == grid_pixel_indexes_util).all()

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)), pixel_scales=(2.0, 4.0)
        )

        grid_arcsec = np.array([[1.0, -2.0], [1.0, 2.0], [-1.0, -2.0], [-1.0, 2.0]])

        grid_pixels_util = aa.util.grid.grid_pixel_indexes_1d_from(
            grid_scaled_1d=grid_arcsec, shape_2d=(2, 2), pixel_scales=(2.0, 4.0)
        )

        grid_pixels = mask.geometry.grid_pixel_indexes_from_grid_scaled_1d(
            grid_scaled_1d=grid_arcsec
        )

        assert (grid_pixels == grid_pixels_util).all()

    def test__grid_arcsec_1d_from_grid_pixels_1d__same_as_grid_util(self):
        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)), pixel_scales=(2.0, 2.0)
        )

        grid_pixels = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])

        grid_pixels_util = aa.util.grid.grid_scaled_1d_from(
            grid_pixels_1d=grid_pixels, shape_2d=(2, 2), pixel_scales=(2.0, 2.0)
        )

        grid_pixels = mask.geometry.grid_scaled_from_grid_pixels_1d(
            grid_pixels_1d=grid_pixels
        )

        assert (grid_pixels == grid_pixels_util).all()

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)), pixel_scales=(2.0, 2.0)
        )

        grid_pixels = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])

        grid_pixels_util = aa.util.grid.grid_scaled_1d_from(
            grid_pixels_1d=grid_pixels, shape_2d=(2, 2), pixel_scales=(2.0, 2.0)
        )
        grid_pixels = mask.geometry.grid_scaled_from_grid_pixels_1d(
            grid_pixels_1d=grid_pixels
        )

        assert (grid_pixels == grid_pixels_util).all()

    def test__pixel_grid__grids_with_nonzero_centres__same_as_grid_util(self):
        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)),
            pixel_scales=(2.0, 2.0),
            origin=(1.0, 2.0),
        )

        grid_arcsec = np.array([[1.0, -2.0], [1.0, 2.0], [-1.0, -2.0], [-1.0, 2.0]])

        grid_pixels_util = aa.util.grid.grid_pixels_1d_from(
            grid_scaled_1d=grid_arcsec,
            shape_2d=(2, 2),
            pixel_scales=(2.0, 2.0),
            origin=(1.0, 2.0),
        )
        grid_pixels = mask.geometry.grid_pixels_from_grid_scaled_1d(
            grid_scaled_1d=grid_arcsec
        )
        assert (grid_pixels == grid_pixels_util).all()

        grid_pixels_util = aa.util.grid.grid_pixel_indexes_1d_from(
            grid_scaled_1d=grid_arcsec,
            shape_2d=(2, 2),
            pixel_scales=(2.0, 2.0),
            origin=(1.0, 2.0),
        )
        grid_pixels = mask.geometry.grid_pixel_indexes_from_grid_scaled_1d(
            grid_scaled_1d=grid_arcsec
        )
        assert grid_pixels == pytest.approx(grid_pixels_util, 1e-4)

        grid_pixels_util = aa.util.grid.grid_pixel_centres_1d_from(
            grid_scaled_1d=grid_arcsec,
            shape_2d=(2, 2),
            pixel_scales=(2.0, 2.0),
            origin=(1.0, 2.0),
        )
        grid_pixels = mask.geometry.grid_pixel_centres_from_grid_scaled_1d(
            grid_scaled_1d=grid_arcsec
        )
        assert grid_pixels == pytest.approx(grid_pixels_util, 1e-4)

        grid_pixels = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])

        grid_arcsec_util = aa.util.grid.grid_scaled_1d_from(
            grid_pixels_1d=grid_pixels,
            shape_2d=(2, 2),
            pixel_scales=(2.0, 2.0),
            origin=(1.0, 2.0),
        )

        grid_arcsec = mask.geometry.grid_scaled_from_grid_pixels_1d(
            grid_pixels_1d=grid_pixels
        )

        assert (grid_arcsec == grid_arcsec_util).all()

        grid_arcsec = np.array([[1.0, -2.0], [1.0, 2.0], [-1.0, -2.0], [-1.0, 2.0]])

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)),
            pixel_scales=(2.0, 1.0),
            origin=(1.0, 2.0),
        )

        grid_pixels_util = aa.util.grid.grid_pixels_1d_from(
            grid_scaled_1d=grid_arcsec,
            shape_2d=(2, 2),
            pixel_scales=(2.0, 1.0),
            origin=(1.0, 2.0),
        )
        grid_pixels = mask.geometry.grid_pixels_from_grid_scaled_1d(
            grid_scaled_1d=grid_arcsec
        )
        assert (grid_pixels == grid_pixels_util).all()

        grid_pixels_util = aa.util.grid.grid_pixel_indexes_1d_from(
            grid_scaled_1d=grid_arcsec,
            shape_2d=(2, 2),
            pixel_scales=(2.0, 1.0),
            origin=(1.0, 2.0),
        )
        grid_pixels = mask.geometry.grid_pixel_indexes_from_grid_scaled_1d(
            grid_scaled_1d=grid_arcsec
        )
        assert (grid_pixels == grid_pixels_util).all()

        grid_pixels_util = aa.util.grid.grid_pixel_centres_1d_from(
            grid_scaled_1d=grid_arcsec,
            shape_2d=(2, 2),
            pixel_scales=(2.0, 1.0),
            origin=(1.0, 2.0),
        )
        grid_pixels = mask.geometry.grid_pixel_centres_from_grid_scaled_1d(
            grid_scaled_1d=grid_arcsec
        )
        assert grid_pixels == pytest.approx(grid_pixels_util, 1e-4)

        grid_pixels = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])

        grid_arcsec_util = aa.util.grid.grid_scaled_1d_from(
            grid_pixels_1d=grid_pixels,
            shape_2d=(2, 2),
            pixel_scales=(2.0, 1.0),
            origin=(1.0, 2.0),
        )

        grid_arcsec = mask.geometry.grid_scaled_from_grid_pixels_1d(
            grid_pixels_1d=grid_pixels
        )

        assert (grid_arcsec == grid_arcsec_util).all()


class TestZoomCentreAndOffet:
    def test__odd_sized_false_mask__centre_is_0_0__pixels_from_centre_are_0_0(self):
        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 3)), pixel_scales=(1.0, 1.0)
        )
        assert mask.geometry._zoom_centre == (1.0, 1.0)
        assert mask.geometry._zoom_offset_pixels == (0, 0)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(5, 5)), pixel_scales=(1.0, 1.0)
        )
        assert mask.geometry._zoom_centre == (2.0, 2.0)
        assert mask.geometry._zoom_offset_pixels == (0, 0)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(3, 5)), pixel_scales=(1.0, 1.0)
        )
        assert mask.geometry._zoom_centre == (1.0, 2.0)
        assert mask.geometry._zoom_offset_pixels == (0, 0)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(5, 3)), pixel_scales=(1.0, 1.0)
        )
        assert mask.geometry._zoom_centre == (2.0, 1.0)
        assert mask.geometry._zoom_offset_pixels == (0, 0)

    def test__even_sized_false_mask__centre_is_0_0__pixels_from_centre_are_0_0(self):
        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(4, 4)), pixel_scales=(1.0, 1.0)
        )
        assert mask.geometry._zoom_centre == (1.5, 1.5)
        assert mask.geometry._zoom_offset_pixels == (0, 0)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(6, 6)), pixel_scales=(1.0, 1.0)
        )
        assert mask.geometry._zoom_centre == (2.5, 2.5)
        assert mask.geometry._zoom_offset_pixels == (0, 0)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(4, 6)), pixel_scales=(1.0, 1.0)
        )
        assert mask.geometry._zoom_centre == (1.5, 2.5)
        assert mask.geometry._zoom_offset_pixels == (0, 0)

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(6, 4)), pixel_scales=(1.0, 1.0)
        )
        assert mask.geometry._zoom_centre == (2.5, 1.5)
        assert mask.geometry._zoom_offset_pixels == (0, 0)

    def test__mask_is_single_false__extraction_centre_is_central_pixel(self):
        mask = aa.Mask.manual(
            mask=np.array(
                [[False, True, True], [True, True, True], [True, True, True]]
            ),
            pixel_scales=(1.0, 1.0),
        )
        assert mask.geometry._zoom_centre == (0, 0)
        assert mask.geometry._zoom_offset_pixels == (-1, -1)

        mask = aa.Mask.manual(
            mask=np.array(
                [[True, True, False], [True, True, True], [True, True, True]]
            ),
            pixel_scales=(1.0, 1.0),
        )
        assert mask.geometry._zoom_centre == (0, 2)
        assert mask.geometry._zoom_offset_pixels == (-1, 1)

        mask = aa.Mask.manual(
            mask=np.array(
                [[True, True, True], [True, True, True], [False, True, True]]
            ),
            pixel_scales=(1.0, 1.0),
        )
        assert mask.geometry._zoom_centre == (2, 0)
        assert mask.geometry._zoom_offset_pixels == (1, -1)

        mask = aa.Mask.manual(
            mask=np.array(
                [[True, True, True], [True, True, True], [True, True, False]]
            ),
            pixel_scales=(1.0, 1.0),
        )
        assert mask.geometry._zoom_centre == (2, 2)
        assert mask.geometry._zoom_offset_pixels == (1, 1)

        mask = aa.Mask.manual(
            mask=np.array(
                [[True, False, True], [True, True, True], [True, True, True]]
            ),
            pixel_scales=(1.0, 1.0),
        )
        assert mask.geometry._zoom_centre == (0, 1)
        assert mask.geometry._zoom_offset_pixels == (-1, 0)

        mask = aa.Mask.manual(
            mask=np.array(
                [[True, True, True], [False, True, True], [True, True, True]]
            ),
            pixel_scales=(1.0, 1.0),
        )
        assert mask.geometry._zoom_centre == (1, 0)
        assert mask.geometry._zoom_offset_pixels == (0, -1)

        mask = aa.Mask.manual(
            mask=np.array(
                [[True, True, True], [True, True, False], [True, True, True]]
            ),
            pixel_scales=(1.0, 1.0),
        )
        assert mask.geometry._zoom_centre == (1, 2)
        assert mask.geometry._zoom_offset_pixels == (0, 1)

        mask = aa.Mask.manual(
            mask=np.array(
                [[True, True, True], [True, True, True], [True, False, True]]
            ),
            pixel_scales=(1.0, 1.0),
        )
        assert mask.geometry._zoom_centre == (2, 1)
        assert mask.geometry._zoom_offset_pixels == (1, 0)

    def test__mask_is_x2_false__extraction_centre_is_central_pixel(self):
        mask = aa.Mask.manual(
            mask=np.array(
                [[False, True, True], [True, True, True], [True, True, False]]
            ),
            pixel_scales=(1.0, 1.0),
        )
        assert mask.geometry._zoom_centre == (1, 1)
        assert mask.geometry._zoom_offset_pixels == (0, 0)

        mask = aa.Mask.manual(
            mask=np.array(
                [[False, True, True], [True, True, True], [False, True, True]]
            ),
            pixel_scales=(1.0, 1.0),
        )
        assert mask.geometry._zoom_centre == (1, 0)
        assert mask.geometry._zoom_offset_pixels == (0, -1)

        mask = aa.Mask.manual(
            mask=np.array(
                [[False, True, False], [True, True, True], [True, True, True]]
            ),
            pixel_scales=(1.0, 1.0),
        )
        assert mask.geometry._zoom_centre == (0, 1)
        assert mask.geometry._zoom_offset_pixels == (-1, 0)

        mask = aa.Mask.manual(
            mask=np.array(
                [[False, False, True], [True, True, True], [True, True, True]]
            ),
            pixel_scales=(1.0, 1.0),
        )
        assert mask.geometry._zoom_centre == (0, 0.5)
        assert mask.geometry._zoom_offset_pixels == (-1, -0.5)

    def test__rectangular_mask(self):
        mask = aa.Mask.manual(
            mask=np.array(
                [
                    [False, True, True, True],
                    [True, True, True, True],
                    [True, True, True, True],
                ]
            ),
            pixel_scales=(1.0, 1.0),
        )

        assert mask.geometry._zoom_centre == (0, 0)
        assert mask.geometry._zoom_offset_pixels == (-1.0, -1.5)

        mask = aa.Mask.manual(
            mask=np.array(
                [
                    [True, True, True, True],
                    [True, True, True, True],
                    [True, True, True, False],
                ]
            ),
            pixel_scales=(1.0, 1.0),
        )

        assert mask.geometry._zoom_centre == (2, 3)
        assert mask.geometry._zoom_offset_pixels == (1.0, 1.5)

        mask = aa.Mask.manual(
            mask=np.array(
                [
                    [True, True, True, True, True],
                    [True, True, True, True, True],
                    [True, True, True, True, False],
                ]
            ),
            pixel_scales=(1.0, 1.0),
        )

        assert mask.geometry._zoom_centre == (2, 4)
        assert mask.geometry._zoom_offset_pixels == (1, 2)

        mask = aa.Mask.manual(
            mask=np.array(
                [
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, False],
                ]
            ),
            pixel_scales=(1.0, 1.0),
        )

        assert mask.geometry._zoom_centre == (2, 6)
        assert mask.geometry._zoom_offset_pixels == (1, 3)

        mask = aa.Mask.manual(
            mask=np.array(
                [
                    [True, True, True],
                    [True, True, True],
                    [True, True, True],
                    [True, True, True],
                    [True, True, False],
                ]
            ),
            pixel_scales=(1.0, 1.0),
        )

        assert mask.geometry._zoom_centre == (4, 2)
        assert mask.geometry._zoom_offset_pixels == (2, 1)

        mask = aa.Mask.manual(
            mask=np.array(
                [
                    [True, True, True],
                    [True, True, True],
                    [True, True, True],
                    [True, True, True],
                    [True, True, True],
                    [True, True, True],
                    [True, True, False],
                ]
            ),
            pixel_scales=(1.0, 1.0),
        )

        assert mask.geometry._zoom_centre == (6, 2)
        assert mask.geometry._zoom_offset_pixels == (3, 1)



import os
import numpy as np

import autoarray as aa
from test_autoarray.mock import (
    MockGridLikeIteratorObj,
    MockGridLikeObj,
    ndarray_1d_from_grid,
    ndarray_2d_from_grid,
)

test_coordinates_dir = "{}/files/coordinates/".format(
    os.path.dirname(os.path.realpath(__file__))
)


def test__grid_in__output_values_same_format():
    mask = np.array(
        [
            [True, True, True, True],
            [True, False, False, True],
            [True, False, False, True],
            [True, True, True, True],
        ]
    )

    mask = aa.Mask.manual(mask=mask, pixel_scales=(1.0, 1.0), sub_size=1)

    grid = aa.Grid.from_mask(mask=mask)

    grid_like_object = MockGridLikeObj()

    array_output = grid_like_object.ndarray_1d_from_grid(grid=grid)

    assert isinstance(array_output, aa.Array)
    assert (
        array_output.in_2d
        == np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ]
        )
    ).all()

    grid_output = grid_like_object.ndarray_2d_from_grid(grid=grid)

    assert isinstance(grid_output, aa.Grid)
    assert (
        grid_output.in_2d
        == np.array(
            [
                [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]],
                [[0.0, 0.0], [1.0, -1.0], [1.0, 1.0], [0.0, 0.0]],
                [[0.0, 0.0], [-1.0, -1.0], [-1.0, 1.0], [0.0, 0.0]],
                [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]],
            ]
        )
    ).all()


def test__grid_in__output_is_list__list_of_same_format():
    mask = np.array(
        [
            [True, True, True, True],
            [True, False, False, True],
            [True, False, False, True],
            [True, True, True, True],
        ]
    )

    mask = aa.Mask.manual(mask=mask, pixel_scales=(1.0, 1.0), sub_size=1)

    grid = aa.Grid.from_mask(mask=mask)

    grid_like_object = MockGridLikeObj()

    array_output = grid_like_object.ndarray_1d_list_from_grid(grid=grid)

    assert isinstance(array_output[0], aa.Array)
    assert (
        array_output[0].in_2d
        == np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 1.0, 0.0],
                [0.0, 1.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ]
        )
    ).all()

    assert isinstance(array_output[1], aa.Array)
    assert (
        array_output[1].in_2d
        == np.array(
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 2.0, 2.0, 0.0],
                [0.0, 2.0, 2.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ]
        )
    ).all()

    grid_output = grid_like_object.ndarray_2d_list_from_grid(grid=grid)

    assert isinstance(grid_output[0], aa.Grid)
    assert (
        grid_output[0].in_2d
        == np.array(
            [
                [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]],
                [[0.0, 0.0], [0.5, -0.5], [0.5, 0.5], [0.0, 0.0]],
                [[0.0, 0.0], [-0.5, -0.5], [-0.5, 0.5], [0.0, 0.0]],
                [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]],
            ]
        )
    ).all()

    assert isinstance(grid_output[1], aa.Grid)
    assert (
        grid_output[1].in_2d
        == np.array(
            [
                [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]],
                [[0.0, 0.0], [1.0, -1.0], [1.0, 1.0], [0.0, 0.0]],
                [[0.0, 0.0], [-1.0, -1.0], [-1.0, 1.0], [0.0, 0.0]],
                [[0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]],
            ]
        )
    ).all()


def test__grid_coordinates_in__output_values_same_format():

    grid_like_object = MockGridLikeObj()

    coordinates = aa.GridCoordinates(
        coordinates=[[(1.0, 2.0), (3.0, 4.0)], [(5.0, 6.0)]]
    )

    values_output = grid_like_object.ndarray_1d_from_grid(grid=coordinates)

    assert values_output.in_list == [[1.0, 1.0], [1.0]]

    coordinates_output = grid_like_object.ndarray_2d_from_grid(grid=coordinates)

    assert coordinates_output.in_list == [[(2.0, 4.0), (6.0, 8.0)], [(10.0, 12.0)]]


def test__grid_coordinates_in__output_is_list__list_of_same_format():

    grid_like_object = MockGridLikeObj()

    coordinates = aa.GridCoordinates(
        coordinates=[[(1.0, 2.0), (3.0, 4.0)], [(5.0, 6.0)]]
    )

    coordinates_output = grid_like_object.ndarray_1d_list_from_grid(grid=coordinates)

    assert coordinates_output[0].in_list == [[1.0, 1.0], [1.0]], [[2.0, 2.0], [2.0]]

    coordinates_output = grid_like_object.ndarray_2d_list_from_grid(grid=coordinates)

    assert coordinates_output[0].in_list == [[(1.0, 2.0), (3.0, 4.0)], [(5.0, 6.0)]], [
        [(2.0, 4.0), (6.0, 8.0)],
        [(10.0, 12.0)],
    ]


def test__grid_iterate_in__output_values__use_iterated_array_function():

    mask = aa.Mask.manual(
        mask=[
            [True, True, True, True, True],
            [True, False, False, False, True],
            [True, False, False, False, True],
            [True, False, False, False, True],
            [True, True, True, True, True],
        ],
        pixel_scales=(1.0, 1.0),
        origin=(0.001, 0.001),
    )

    grid = aa.GridIterate.from_mask(
        mask=mask, fractional_accuracy=1.0, sub_steps=[2, 3]
    )

    grid_like_obj = MockGridLikeIteratorObj()

    values = grid_like_obj.ndarray_1d_from_grid(grid=grid)

    mask_sub_3 = mask.mask_new_sub_size_from_mask(mask=mask, sub_size=3)
    grid_sub_3 = aa.Grid.from_mask(mask=mask_sub_3)
    values_sub_3 = ndarray_1d_from_grid(grid=grid_sub_3, profile=None)
    values_sub_3 = grid_sub_3.structure_from_result(result=values_sub_3)

    assert (values == values_sub_3.in_1d_binned).all()

    grid = aa.GridIterate.from_mask(
        mask=mask, fractional_accuracy=0.000001, sub_steps=[2, 4, 8, 16, 32]
    )

    grid_like_obj = MockGridLikeIteratorObj()

    values = grid_like_obj.ndarray_1d_from_grid(grid=grid)

    mask_sub_2 = mask.mask_new_sub_size_from_mask(mask=mask, sub_size=2)
    grid_sub_2 = aa.Grid.from_mask(mask=mask_sub_2)
    values_sub_2 = ndarray_1d_from_grid(grid=grid_sub_2, profile=None)
    values_sub_2 = grid_sub_2.structure_from_result(result=values_sub_2)

    assert (values == values_sub_2.in_1d_binned).all()

    grid = aa.GridIterate.from_mask(
        mask=mask, fractional_accuracy=0.5, sub_steps=[2, 4]
    )

    iterate_obj = MockGridLikeIteratorObj()

    values = iterate_obj.ndarray_1d_from_grid(grid=grid)

    mask_sub_2 = mask.mask_new_sub_size_from_mask(mask=mask, sub_size=2)
    grid_sub_2 = aa.Grid.from_mask(mask=mask_sub_2)
    values_sub_2 = ndarray_1d_from_grid(grid=grid_sub_2, profile=None)
    values_sub_2 = grid_sub_2.structure_from_result(result=values_sub_2)

    mask_sub_4 = mask.mask_new_sub_size_from_mask(mask=mask, sub_size=4)
    grid_sub_4 = aa.Grid.from_mask(mask=mask_sub_4)
    values_sub_4 = ndarray_1d_from_grid(grid=grid_sub_4, profile=None)
    values_sub_4 = grid_sub_4.structure_from_result(result=values_sub_4)

    assert values.in_2d[1, 1] == values_sub_2.in_2d_binned[1, 1]
    assert values.in_2d[2, 2] != values_sub_2.in_2d_binned[2, 2]

    assert values.in_2d[1, 1] != values_sub_4.in_2d_binned[1, 1]
    assert values.in_2d[2, 2] == values_sub_4.in_2d_binned[2, 2]


def test__grid_iterate_in__output_is_list_of_arrays__use_maximum_sub_size_in_all_pixels():

    mask = aa.Mask.manual(
        mask=[
            [True, True, True, True, True],
            [True, False, False, False, True],
            [True, False, False, False, True],
            [True, False, False, False, True],
            [True, True, True, True, True],
        ],
        pixel_scales=(1.0, 1.0),
        origin=(0.001, 0.001),
    )

    grid = aa.GridIterate.from_mask(
        mask=mask, fractional_accuracy=0.05, sub_steps=[2, 3]
    )

    grid_like_obj = MockGridLikeIteratorObj()

    values = grid_like_obj.ndarray_1d_list_from_grid(grid=grid)

    mask_sub_3 = mask.mask_new_sub_size_from_mask(mask=mask, sub_size=3)
    grid_sub_3 = aa.Grid.from_mask(mask=mask_sub_3)
    values_sub_3 = ndarray_1d_from_grid(grid=grid_sub_3, profile=None)
    values_sub_3 = grid_sub_3.structure_from_result(result=values_sub_3)

    assert (values[0] == values_sub_3.in_1d_binned).all()


def test__grid_iterate_in__output_values__use_iterated_grid_function():

    mask = aa.Mask.manual(
        mask=[
            [True, True, True, True, True],
            [True, False, False, False, True],
            [True, False, False, False, True],
            [True, False, False, False, True],
            [True, True, True, True, True],
        ],
        pixel_scales=(1.0, 1.0),
        origin=(0.001, 0.001),
    )

    grid = aa.GridIterate.from_mask(
        mask=mask, fractional_accuracy=1.0, sub_steps=[2, 3]
    )

    grid_like_obj = MockGridLikeIteratorObj()

    values = grid_like_obj.ndarray_2d_from_grid(grid=grid)

    mask_sub_3 = mask.mask_new_sub_size_from_mask(mask=mask, sub_size=3)
    grid_sub_3 = aa.Grid.from_mask(mask=mask_sub_3)
    values_sub_3 = ndarray_2d_from_grid(grid=grid_sub_3, profile=None)
    values_sub_3 = grid_sub_3.structure_from_result(result=values_sub_3)

    assert (values == values_sub_3.in_1d_binned).all()

    grid = aa.GridIterate.from_mask(
        mask=mask, fractional_accuracy=0.000001, sub_steps=[2, 4, 8, 16, 32]
    )

    grid_like_obj = MockGridLikeIteratorObj()

    values = grid_like_obj.ndarray_2d_from_grid(grid=grid)

    mask_sub_2 = mask.mask_new_sub_size_from_mask(mask=mask, sub_size=2)
    grid_sub_2 = aa.Grid.from_mask(mask=mask_sub_2)
    values_sub_2 = ndarray_2d_from_grid(grid=grid_sub_2, profile=None)
    values_sub_2 = grid_sub_2.structure_from_result(result=values_sub_2)

    assert (values == values_sub_2.in_1d_binned).all()

    grid = aa.GridIterate.from_mask(
        mask=mask, fractional_accuracy=0.5, sub_steps=[2, 4]
    )

    iterate_obj = MockGridLikeIteratorObj()

    values = iterate_obj.ndarray_2d_from_grid(grid=grid)

    mask_sub_2 = mask.mask_new_sub_size_from_mask(mask=mask, sub_size=2)
    grid_sub_2 = aa.Grid.from_mask(mask=mask_sub_2)
    values_sub_2 = ndarray_2d_from_grid(grid=grid_sub_2, profile=None)
    values_sub_2 = grid_sub_2.structure_from_result(result=values_sub_2)

    mask_sub_4 = mask.mask_new_sub_size_from_mask(mask=mask, sub_size=4)
    grid_sub_4 = aa.Grid.from_mask(mask=mask_sub_4)
    values_sub_4 = ndarray_2d_from_grid(grid=grid_sub_4, profile=None)
    values_sub_4 = grid_sub_4.structure_from_result(result=values_sub_4)

    assert values.in_2d[1, 1, 0] == values_sub_2.in_2d_binned[1, 1, 0]
    assert values.in_2d[2, 2, 0] != values_sub_2.in_2d_binned[2, 2, 0]

    assert values.in_2d[1, 1, 0] != values_sub_4.in_2d_binned[1, 1, 0]
    assert values.in_2d[2, 2, 0] == values_sub_4.in_2d_binned[2, 2, 0]

    assert values.in_2d[1, 1, 1] == values_sub_2.in_2d_binned[1, 1, 1]
    assert values.in_2d[2, 2, 1] != values_sub_2.in_2d_binned[2, 2, 1]

    assert values.in_2d[1, 1, 1] != values_sub_4.in_2d_binned[1, 1, 1]
    assert values.in_2d[2, 2, 1] == values_sub_4.in_2d_binned[2, 2, 1]


def test__grid_iterate_in__output_is_list_of_grids__use_maximum_sub_size_in_all_pixels():

    mask = aa.Mask.manual(
        mask=[
            [True, True, True, True, True],
            [True, False, False, False, True],
            [True, False, False, False, True],
            [True, False, False, False, True],
            [True, True, True, True, True],
        ],
        pixel_scales=(1.0, 1.0),
        origin=(0.001, 0.001),
    )

    grid = aa.GridIterate.from_mask(
        mask=mask, fractional_accuracy=0.05, sub_steps=[2, 3]
    )

    grid_like_obj = MockGridLikeIteratorObj()

    values = grid_like_obj.ndarray_2d_list_from_grid(grid=grid)

    mask_sub_3 = mask.mask_new_sub_size_from_mask(mask=mask, sub_size=3)
    grid_sub_3 = aa.Grid.from_mask(mask=mask_sub_3)
    values_sub_3 = ndarray_2d_from_grid(grid=grid_sub_3, profile=None)
    values_sub_3 = grid_sub_3.structure_from_result(result=values_sub_3)

    assert (values[0][0] == values_sub_3.in_1d_binned[0]).all()
    assert (values[0][1] == values_sub_3.in_1d_binned[1]).all()


def test__grid_interpolate_in__output_values__interpolation_used_and_accurate():

    mask = aa.Mask.circular_annular(
        shape_2d=(20, 20),
        pixel_scales=(1.0, 1.0),
        sub_size=1,
        inner_radius=3.0,
        outer_radius=8.0,
    )

    grid_like_obj = MockGridLikeIteratorObj()

    grid = aa.Grid.from_mask(mask=mask)

    true_array = grid_like_obj.ndarray_1d_from_grid(grid=grid)

    grid = aa.GridInterpolate.from_mask(mask=mask, pixel_scales_interp=1.0)

    interpolated_array = grid_like_obj.ndarray_1d_from_grid(grid=grid)

    assert interpolated_array.shape[0] == mask.pixels_in_mask
    assert (true_array == interpolated_array).all()

    grid = aa.GridInterpolate.from_mask(mask=mask, pixel_scales_interp=0.1)

    interpolated_array = grid_like_obj.ndarray_1d_from_grid(grid=grid)

    assert interpolated_array.shape[0] == mask.pixels_in_mask
    assert true_array[0] != interpolated_array[0]
    assert np.max(true_array - interpolated_array) < 0.001

    grid = aa.Grid.from_mask(mask=mask)

    true_grid = grid_like_obj.ndarray_2d_from_grid(grid=grid)

    grid = aa.GridInterpolate.from_mask(mask=mask, pixel_scales_interp=1.0)

    interpolated_grid = grid_like_obj.ndarray_2d_from_grid(grid=grid)

    assert interpolated_grid.shape[0] == mask.pixels_in_mask
    assert interpolated_grid.shape[1] == 2
    assert (true_grid == interpolated_grid).all()

    grid = aa.GridInterpolate.from_mask(mask=mask, pixel_scales_interp=0.1)

    interpolated_grid = grid_like_obj.ndarray_2d_from_grid(grid=grid)

    assert interpolated_grid.shape[0] == mask.pixels_in_mask
    assert interpolated_grid.shape[1] == 2
    assert true_grid[0, 0] != interpolated_grid[0, 0]
    assert np.max(true_grid[:, 0] - interpolated_grid[:, 0]) < 0.001
    assert np.max(true_grid[:, 1] - interpolated_grid[:, 1]) < 0.001



import autoarray as aa

import numpy as np
import pytest


class MockRealSpaceMask:
    def __init__(self, grid):

        self.grid = grid
        self.geometry = MockGeometry(grid=grid)

    @property
    def mask_sub_1(self):
        return self

    @property
    def pixels_in_mask(self):
        return self.geometry.masked_grid.in_1d_binned.in_radians.shape[0]

    @property
    def pixel_scales(self):
        return self.grid.pixel_scales

    @property
    def sub_size(self):
        return self.grid.sub_size

    @property
    def origin(self):
        return self.grid.origin


class MockGeometry:
    def __init__(self, grid):

        self.masked_grid = MockMaskedGrid(grid=grid)


class MockMaskedGrid:
    def __init__(self, grid):
        self.in_1d_binned = MockMaskedGrid2(grid=grid)


class MockMaskedGrid2:
    def __init__(self, grid):
        self.in_radians = grid


class TestVisiblities:
    def test__real_visibilities__intensity_image_all_ones__simple_cases(self):

        uv_wavelengths = np.ones(shape=(4, 2))

        grid_radians = aa.Grid.manual_2d([[[1.0, 1.0]]], pixel_scales=1.0)

        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        image = aa.Array.ones(shape_2d=(1, 1))

        real_visibilities = transformer.real_visibilities_from_image(image=image)

        assert (real_visibilities == np.ones(shape=4)).all()

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])

        grid_radians = aa.Grid.manual_2d([[[0.1, 0.2], [0.3, 0.4]]], pixel_scales=1.0)

        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        image = aa.Array.ones(shape_2d=(1, 2))

        real_visibilities = transformer.real_visibilities_from_image(image=image)

        assert real_visibilities == pytest.approx(
            np.array([-0.091544, -0.73359736, -0.613160]), 1.0e-4
        )

    def test__real_visibilities__intensity_image_varies__simple_cases(self):

        uv_wavelengths = np.ones(shape=(4, 2))
        grid_radians = aa.Grid.manual_2d([[[1.0, 1.0]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        image = aa.Array.manual_2d([[2.0]])

        real_visibilities = transformer.real_visibilities_from_image(image=image)

        assert (real_visibilities == np.array([2.0])).all()

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])

        grid_radians = aa.Grid.manual_2d([[[0.1, 0.2], [0.3, 0.4]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        image = aa.Array.manual_2d([[3.0, 6.0]])

        real_visibilities = transformer.real_visibilities_from_image(image=image)

        assert real_visibilities == pytest.approx(
            np.array([-2.46153, -5.14765, -3.11681]), 1.0e-4
        )

    def test__real_visibilities__preload_and_non_preload_give_same_answer(self):

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])
        grid_radians = aa.Grid.manual_2d([[[0.1, 0.2], [0.3, 0.4]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer_preload = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=True,
        )
        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        image = aa.Array.manual_2d([[2.0, 6.0]])

        real_visibilities_via_preload = transformer_preload.real_visibilities_from_image(
            image=image
        )
        real_visibilities = transformer.real_visibilities_from_image(image=image)

        assert (real_visibilities_via_preload == real_visibilities).all()

    def test__imag_visibilities__intensity_image_all_ones__simple_cases(self):

        uv_wavelengths = np.ones(shape=(4, 2))
        grid_radians = aa.Grid.manual_2d([[[1.0, 1.0]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        image = aa.Array.ones(shape_2d=(1, 1))

        imag_visibilities = transformer.imag_visibilities_from_image(image=image)

        assert imag_visibilities == pytest.approx(np.zeros(shape=4), 1.0e-4)

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])

        grid_radians = aa.Grid.manual_2d([[[0.1, 0.2], [0.3, 0.4]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        image = aa.Array.ones(shape_2d=(2, 1))

        imag_visibilities = transformer.imag_visibilities_from_image(image=image)

        assert imag_visibilities == pytest.approx(
            np.array([-1.45506, -0.781201, -0.077460]), 1.0e-4
        )

    def test__imag_visibilities__intensity_image_varies__simple_cases(self):

        uv_wavelengths = np.ones(shape=(4, 2))
        grid_radians = aa.Grid.manual_2d([[[1.0, 1.0]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        image = aa.Array.manual_2d([[2.0]])

        imag_visibilities = transformer.imag_visibilities_from_image(image=image)

        assert imag_visibilities == pytest.approx(np.zeros((4,)), 1.0e-4)

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])

        grid_radians = aa.Grid.manual_2d([[[0.1, 0.2], [0.3, 0.4]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        image = aa.Array.manual_2d([[3.0, 6.0]])

        imag_visibilities = transformer.imag_visibilities_from_image(image=image)

        assert imag_visibilities == pytest.approx(
            np.array([-6.418822, -1.78146, 2.48210]), 1.0e-4
        )

    def test__imag_visibilities__preload_and_non_preload_give_same_answer(self):

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])
        grid_radians = aa.Grid.manual_2d([[[0.1, 0.2], [0.3, 0.4]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer_preload = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=True,
        )
        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        image = aa.Array.manual_2d([[2.0, 6.0]])

        imag_visibilities_via_preload = transformer_preload.imag_visibilities_from_image(
            image=image
        )
        imag_visibilities = transformer.imag_visibilities_from_image(image=image)

        assert (imag_visibilities_via_preload == imag_visibilities).all()

    def test__visiblities_from_image__same_as_individual_calculations_above(self):

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])

        grid_radians = aa.Grid.manual_2d([[[0.1, 0.2], [0.3, 0.4]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        image = aa.Array.manual_2d([[3.0, 6.0]])

        visibilities = transformer.visibilities_from_image(image=image)

        assert visibilities[:, 0] == pytest.approx(
            np.array([-2.46153, -5.14765, -3.11681]), 1.0e-4
        )
        assert visibilities[:, 1] == pytest.approx(
            np.array([-6.418822, -1.78146, 2.48210]), 1.0e-4
        )

        real_visibilities = transformer.real_visibilities_from_image(image=image)
        imag_visibilities = transformer.imag_visibilities_from_image(image=image)

        assert (visibilities[:, 0] == real_visibilities).all()
        assert (visibilities[:, 1] == imag_visibilities).all()


class TestVisiblitiesMappingMatrix:
    def test__real_visibilities__mapping_matrix_all_ones__simple_cases(self):

        uv_wavelengths = np.ones(shape=(4, 2))
        grid_radians = aa.Grid.manual_2d([[[1.0, 1.0]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        mapping_matrix = np.ones(shape=(1, 1))

        transformed_mapping_matrix = transformer.real_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert (transformed_mapping_matrix == np.ones(shape=(4, 1))).all()

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])

        grid_radians = aa.Grid.manual_2d([[[0.1, 0.2], [0.3, 0.4]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        mapping_matrix = np.ones(shape=(2, 1))

        transformed_mapping_matrix = transformer.real_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert transformed_mapping_matrix == pytest.approx(
            np.array([[-0.091544], [-0.733597], [-0.613160]]), 1.0e-4
        )

        mapping_matrix = np.ones(shape=(2, 2))

        transformed_mapping_matrix = transformer.real_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert transformed_mapping_matrix == pytest.approx(
            np.array(
                [[-0.091544, -0.091544], [-0.733597, -0.733597], [-0.61316, -0.61316]]
            ),
            1.0e-4,
        )

    def test__real_visibilities__more_complex_mapping_matrix(self):

        grid_radians = aa.Grid.manual_2d(
            [[[0.1, 0.2], [0.3, 0.4], [0.5, 0.6]]], pixel_scales=1.0
        )
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        uv_wavelengths = np.array([[0.7, 0.8], [0.9, 1.0]])

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        mapping_matrix = np.array([[1.0], [0.0], [0.0]])

        transformed_mapping_matrix = transformer.real_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert transformed_mapping_matrix == pytest.approx(
            np.array([[0.18738], [-0.18738]]), 1.0e-4
        )

        mapping_matrix = np.array([[0.0], [1.0], [0.0]])

        transformed_mapping_matrix = transformer.real_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert transformed_mapping_matrix == pytest.approx(
            np.array([[-0.992111], [-0.53582]]), 1.0e-4
        )

        mapping_matrix = np.array([[0.0, 0.5], [0.0, 0.2], [1.0, 0.0]])

        transformed_mapping_matrix = transformer.real_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert transformed_mapping_matrix == pytest.approx(
            np.array([[0.42577, -0.10473], [0.968583, -0.20085]]), 1.0e-4
        )

    def test__real_visibilities__preload_and_non_preload_give_same_answer(self):

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])
        grid_radians = aa.Grid.manual_2d([[[0.1, 0.2], [0.3, 0.4]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer_preload = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=True,
        )

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        mapping_matrix = np.array([[3.0, 5.0], [1.0, 2.0]])

        transformed_mapping_matrix_preload = transformer_preload.real_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        transformed_mapping_matrix = transformer.real_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert (transformed_mapping_matrix_preload == transformed_mapping_matrix).all()

    def test__imag_visibilities__mapping_matrix_all_ones__simple_cases(self):

        uv_wavelengths = np.ones(shape=(4, 2))
        grid_radians = aa.Grid.manual_2d([[[1.0, 1.0]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        mapping_matrix = np.ones(shape=(1, 1))

        transformed_mapping_matrix = transformer.imag_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert transformed_mapping_matrix == pytest.approx(
            np.zeros(shape=(4, 1)), 1.0e-4
        )

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])
        grid_radians = aa.Grid.manual_2d([[[0.1, 0.2], [0.3, 0.4]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        mapping_matrix = np.ones(shape=(2, 1))

        transformed_mapping_matrix = transformer.imag_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert transformed_mapping_matrix == pytest.approx(
            np.array([[-1.455060], [-0.78120], [-0.07746]]), 1.0e-4
        )

        mapping_matrix = np.ones(shape=(2, 2))

        transformed_mapping_matrix = transformer.imag_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert transformed_mapping_matrix == pytest.approx(
            np.array(
                [[-1.45506, -1.45506], [-0.78120, -0.78120], [-0.07746, -0.07746]]
            ),
            1.0e-4,
        )

    def test__imag_visibilities__more_complex_mapping_matrix(self):

        grid_radians = aa.Grid.manual_2d(
            [[[0.1, 0.2], [0.3, 0.4], [0.5, 0.6]]], pixel_scales=1.0
        )
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        uv_wavelengths = np.array([[0.7, 0.8], [0.9, 1.0]])

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        mapping_matrix = np.array([[1.0], [0.0], [0.0]])

        transformed_mapping_matrix = transformer.imag_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert transformed_mapping_matrix == pytest.approx(
            np.array([[-0.982287], [-0.982287]]), 1.0e-4
        )

        mapping_matrix = np.array([[0.0], [1.0], [0.0]])

        transformed_mapping_matrix = transformer.imag_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert transformed_mapping_matrix == pytest.approx(
            np.array([[0.12533], [0.84432]]), 1.0e-4
        )

        mapping_matrix = np.array([[0.0, 0.5], [0.0, 0.2], [1.0, 0.0]])

        transformed_mapping_matrix = transformer.imag_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert transformed_mapping_matrix == pytest.approx(
            np.array([[0.90482, -0.46607], [-0.24868, -0.32227]]), 1.0e-4
        )

    def test__imag_visibilities__preload_and_non_preload_give_same_answer(self):

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])
        grid_radians = aa.Grid.manual_2d([[[0.1, 0.2], [0.3, 0.4]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer_preload = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=True,
        )

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        mapping_matrix = np.array([[3.0, 5.0], [1.0, 2.0]])

        transformed_mapping_matrix_preload = transformer_preload.imag_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        transformed_mapping_matrix = transformer.imag_transformed_mapping_matrix_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert (transformed_mapping_matrix_preload == transformed_mapping_matrix).all()

    def test__transformed_mapping_matrices_from_mapping_matrix__same_as_individual_calculations_above(
        self
    ):

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])
        grid_radians = aa.Grid.manual_2d([[[0.1, 0.2], [0.3, 0.4]]], pixel_scales=1.0)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        transformer = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        mapping_matrix = np.ones(shape=(2, 1))

        transformed_mapping_matrices = transformer.transformed_mapping_matrices_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert transformed_mapping_matrices[0] == pytest.approx(
            np.array([[-0.09154], [-0.73359], [-0.61316]]), 1.0e-4
        )
        assert transformed_mapping_matrices[1] == pytest.approx(
            np.array([[-1.45506], [-0.78120], [-0.07746]]), 1.0e-4
        )


class TestTransformerNUFFT:
    def test__image_from_visibilities__same_as_direct__include_numerics(self):

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])

        grid_radians = aa.Grid.uniform(shape_2d=(5, 5), pixel_scales=0.005).in_radians
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        image = aa.Array.ones(
            shape_2d=grid_radians.shape_2d, pixel_scales=grid_radians.pixel_scales
        )

        transformer_dft = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        visibilities_dft = transformer_dft.visibilities_from_image(image=image.in_2d)

        real_space_mask = aa.Mask.unmasked(shape_2d=(5, 5), pixel_scales=0.005)

        transformer_nufft = aa.TransformerNUFFT(
            uv_wavelengths=uv_wavelengths, real_space_mask=real_space_mask
        )

        visibilities_nufft = transformer_nufft.visibilities_from_image(
            image=image.in_2d
        )

        assert visibilities_dft == pytest.approx(visibilities_nufft, 1.0e-2)

        assert visibilities_nufft[0, 0] == pytest.approx(25.02317617953263, 1.0e-7)

    def test__mapping_matix_from_visibilities__same_as_direct__include_numerics(self):

        uv_wavelengths = np.array([[0.2, 1.0], [0.5, 1.1], [0.8, 1.2]])

        grid_radians = aa.Grid.uniform(shape_2d=(5, 5), pixel_scales=0.005)
        real_space_mask = MockRealSpaceMask(grid=grid_radians)

        mapping_matrix = np.ones(shape=(25, 3))

        transformer_dft = aa.TransformerDFT(
            uv_wavelengths=uv_wavelengths,
            real_space_mask=real_space_mask,
            preload_transform=False,
        )

        transformed_mapping_matrices_dft = transformer_dft.transformed_mapping_matrices_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        real_space_mask = aa.Mask.unmasked(shape_2d=(5, 5), pixel_scales=0.005)

        transformer_nufft = aa.TransformerNUFFT(
            uv_wavelengths=uv_wavelengths, real_space_mask=real_space_mask
        )

        transformed_mapping_matrices_nufft = transformer_nufft.transformed_mapping_matrices_from_mapping_matrix(
            mapping_matrix=mapping_matrix
        )

        assert transformed_mapping_matrices_dft[0] == pytest.approx(
            transformed_mapping_matrices_nufft[0], 1.0e-2
        )
        assert transformed_mapping_matrices_dft[1] == pytest.approx(
            transformed_mapping_matrices_nufft[1], 1.0e-2
        )

        assert transformed_mapping_matrices_nufft[0][0, 0] == pytest.approx(
            24.99999636, 1.0e-2
        )
        assert transformed_mapping_matrices_nufft[1][0, 0] == pytest.approx(0.0, 1.0e-2)



from os import path
import os
import pytest

import autoarray as aa
import autoarray.plot as aplt


directory = path.dirname(path.realpath(__file__))


@pytest.fixture(name="plot_path")
def make_imaging_plotter_setup():
    plot_path = "{}/files/plots/imaging/".format(
        os.path.dirname(os.path.realpath(__file__))
    )

    return plot_path


@pytest.fixture(autouse=True)
def set_config_path():
    aa.conf.instance = aa.conf.Config(
        path.join(directory, "files/plotter"), path.join(directory, "output")
    )


def test__individual_attributes_are_output(
    imaging_7x7, positions_7x7, mask_7x7, plot_path, plot_patch
):

    aplt.Imaging.image(
        imaging=imaging_7x7,
        positions=positions_7x7,
        mask=mask_7x7,
        include=aplt.Include(mask=True),
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert plot_path + "image.png" in plot_patch.paths

    aplt.Imaging.noise_map(
        imaging=imaging_7x7,
        mask=mask_7x7,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert plot_path + "noise_map.png" in plot_patch.paths

    aplt.Imaging.psf(
        imaging=imaging_7x7,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert plot_path + "psf.png" in plot_patch.paths

    aplt.Imaging.inverse_noise_map(
        imaging=imaging_7x7,
        mask=mask_7x7,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert plot_path + "inverse_noise_map.png" in plot_patch.paths

    aplt.Imaging.signal_to_noise_map(
        imaging=imaging_7x7,
        mask=mask_7x7,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert plot_path + "signal_to_noise_map.png" in plot_patch.paths


def test__subplot_is_output(
    imaging_7x7, positions_7x7, mask_7x7, plot_path, plot_patch
):

    aplt.Imaging.subplot_imaging(
        imaging=imaging_7x7,
        sub_plotter=aplt.SubPlotter(output=aplt.Output(plot_path, format="png")),
    )

    assert plot_path + "subplot_imaging.png" in plot_patch.paths


def test__imaging_individuals__output_dependent_on_input(
    imaging_7x7, plot_path, plot_patch
):
    aplt.Imaging.individual(
        imaging=imaging_7x7,
        plot_image=True,
        plot_psf=True,
        plot_inverse_noise_map=True,
        plot_absolute_signal_to_noise_map=True,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert plot_path + "image.png" in plot_patch.paths

    assert not plot_path + "noise_map.png" in plot_patch.paths

    assert plot_path + "psf.png" in plot_patch.paths

    assert plot_path + "inverse_noise_map.png" in plot_patch.paths

    assert not plot_path + "signal_to_noise_map.png" in plot_patch.paths

    assert plot_path + "absolute_signal_to_noise_map.png" in plot_patch.paths

    assert not plot_path + "potential_chi_squared_map.png" in plot_patch.paths


def test__output_as_fits__correct_output_format(
    imaging_7x7, positions_7x7, mask_7x7, plot_path, plot_patch
):

    aplt.Imaging.individual(
        imaging=imaging_7x7,
        plot_image=True,
        plot_psf=True,
        plot_absolute_signal_to_noise_map=True,
        plotter=aplt.Plotter(output=aplt.Output(path=plot_path, format="fits")),
    )

    image_from_plot = aa.util.array.numpy_array_2d_from_fits(
        file_path=plot_path + "image.fits", hdu=0
    )

    assert image_from_plot.shape == (7, 7)



from .grid import Grid, GridSparse, GridTransformed, GridTransformedNumpy
from .iterate import GridIterate
from .interpolate import GridInterpolate
from .coordinates import GridCoordinates, GridCoordinatesUniform
from .pixelization import GridRectangular, GridVoronoi
from .decorators import *



import os
import shutil
import numpy as np
import pytest

import autoarray as aa
from autoarray.structures import grids

test_coordinates_dir = "{}/files/coordinates/".format(
    os.path.dirname(os.path.realpath(__file__))
)


class TestAbstractGridCoordinates:
    def test__indexes_give_entries_where_list_begin_and_end(self):

        coordinates = aa.GridCoordinates(coordinates=[[(0.0, 0.0)]])

        assert coordinates.lower_indexes == [0]
        assert coordinates.upper_indexes == [1]

        coordinates = aa.GridCoordinates(coordinates=[[(0.0, 0.0), (0.0, 0.0)]])

        assert coordinates.lower_indexes == [0]
        assert coordinates.upper_indexes == [2]

        coordinates = aa.GridCoordinates(
            coordinates=[[(0.0, 0.0), (0.0, 0.0)], [(0.0, 0.0)]]
        )

        assert coordinates.lower_indexes == [0, 2]
        assert coordinates.upper_indexes == [2, 3]

        coordinates = aa.GridCoordinates(
            coordinates=[
                [(0.0, 0.0), (0.0, 0.0)],
                [(0.0, 0.0), (0.0, 0.0), (0.0, 0.0)],
                [(0.0, 0.0), (0.0, 0.0)],
                [(0.0, 0.0)],
            ]
        )

        assert coordinates.lower_indexes == [0, 2, 5, 7]
        assert coordinates.upper_indexes == [2, 5, 7, 8]

    def test__input_as_list_or_list_of_other_types__all_convert_correctly(self):

        # Input tuples

        coordinates = aa.GridCoordinates(coordinates=[(1.0, -1.0), (1.0, 1.0)])

        assert type(coordinates) == grids.GridCoordinates
        assert (coordinates == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert coordinates.in_list == [[(1.0, -1.0), (1.0, 1.0)]]

        coordinates = aa.GridCoordinates(coordinates=[[(1.0, -1.0), (1.0, 1.0)]])

        assert type(coordinates) == grids.GridCoordinates
        assert (coordinates == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert coordinates.in_list == [[(1.0, -1.0), (1.0, 1.0)]]

        coordinates = aa.GridCoordinates(coordinates=[(1.0, -1.0), (1.0, 1.0)])

        assert type(coordinates) == grids.GridCoordinates
        assert (coordinates == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert coordinates.in_list == [[(1.0, -1.0), (1.0, 1.0)]]

        coordinates = aa.GridCoordinates(coordinates=[[(1.0, -1.0), (1.0, 1.0)]])

        assert type(coordinates) == grids.GridCoordinates
        assert (coordinates == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert coordinates.in_list == [[(1.0, -1.0), (1.0, 1.0)]]

        # Input np array

        coordinates = aa.GridCoordinates(
            coordinates=[np.array([1.0, -1.0]), np.array([1.0, 1.0])]
        )

        assert type(coordinates) == grids.GridCoordinates
        assert (coordinates == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert coordinates.in_list == [[(1.0, -1.0), (1.0, 1.0)]]

        coordinates = aa.GridCoordinates(
            coordinates=[[np.array([1.0, -1.0]), np.array([1.0, 1.0])]]
        )

        assert type(coordinates) == grids.GridCoordinates
        assert (coordinates == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert coordinates.in_list == [[(1.0, -1.0), (1.0, 1.0)]]

        coordinates = aa.GridCoordinates(
            coordinates=[np.array([[1.0, -1.0], [1.0, 1.0]])]
        )

        assert type(coordinates) == grids.GridCoordinates
        assert (coordinates == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert coordinates.in_list == [[(1.0, -1.0), (1.0, 1.0)]]

        # Input list

        coordinates = aa.GridCoordinates(coordinates=[[1.0, -1.0], [1.0, 1.0]])

        assert type(coordinates) == grids.GridCoordinates
        assert (coordinates == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert coordinates.in_list == [[(1.0, -1.0), (1.0, 1.0)]]

        coordinates = aa.GridCoordinates(coordinates=[[[1.0, -1.0]], [[1.0, 1.0]]])

        assert type(coordinates) == grids.GridCoordinates
        assert (coordinates == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert coordinates.in_list == [[(1.0, -1.0)], [(1.0, 1.0)]]

        coordinates = aa.GridCoordinates(coordinates=[[[1.0, -1.0], [1.0, 1.0]]])

        assert type(coordinates) == grids.GridCoordinates
        assert (coordinates == np.array([[[1.0, -1.0], [1.0, 1.0]]])).all()
        assert coordinates.in_list == [[(1.0, -1.0), (1.0, 1.0)]]

    def test__values_from_arr_1d(self):

        coordinates = aa.GridCoordinates(coordinates=[[(1.0, 1.0), (2.0, 2.0)]])

        values_from_1d = coordinates.values_from_arr_1d(arr_1d=np.array([1.0, 2.0]))

        assert values_from_1d.in_list == [[1.0, 2.0]]

        coordinates = aa.GridCoordinates(
            coordinates=[[(1.0, 1.0), (2.0, 2.0)], [(3.0, 3.0)]]
        )

        values_from_1d = coordinates.values_from_arr_1d(
            arr_1d=np.array([1.0, 2.0, 3.0])
        )

        assert values_from_1d.in_list == [[1.0, 2.0], [3.0]]

    def test__with_mask__converts_to_and_from_pixels(self):

        mask = aa.Mask.manual(
            mask=np.full(fill_value=False, shape=(2, 2)), pixel_scales=(2.0, 2.0)
        )

        coordinates = aa.GridCoordinates(coordinates=[[(1.0, -1.0), (1.0, 1.0)]])

        assert coordinates.in_list == [[(1.0, -1.0), (1.0, 1.0)]]

        coordinates = aa.GridCoordinates.from_pixels_and_mask(
            pixels=[[(0, 0), (0, 1)]], mask=mask
        )

        assert coordinates.in_list == [[(1.0, -1.0), (1.0, 1.0)]]

    def test__load_coordinates__retains_list_structure(self):
        coordinates = aa.GridCoordinates.from_file(
            file_path=test_coordinates_dir + "coordinates_test.dat"
        )

        assert coordinates.in_list == [
            [(1.0, 1.0), (2.0, 2.0)],
            [(3.0, 3.0), (4.0, 4.0), (5.0, 6.0)],
        ]

    def test__output_coordinates_to_file(self):
        coordinates = aa.GridCoordinates(
            [[(4.0, 4.0), (5.0, 5.0)], [(6.0, 6.0), (7.0, 7.0), (8.0, 8.0)]]
        )

        output_coordinates_dir = "{}/files/coordinates/output_test/".format(
            os.path.dirname(os.path.realpath(__file__))
        )
        if os.path.exists(output_coordinates_dir):
            shutil.rmtree(output_coordinates_dir)

        os.makedirs(output_coordinates_dir)

        coordinates.output_to_file(
            file_path=output_coordinates_dir + "coordinates_test.dat"
        )

        coordinates = aa.GridCoordinates.from_file(
            file_path=output_coordinates_dir + "coordinates_test.dat"
        )

        assert coordinates.in_list == [
            [(4.0, 4.0), (5.0, 5.0)],
            [(6.0, 6.0), (7.0, 7.0), (8.0, 8.0)],
        ]

        with pytest.raises(FileExistsError):
            coordinates.output_to_file(
                file_path=output_coordinates_dir + "coordinates_test.dat"
            )

        coordinates.output_to_file(
            file_path=output_coordinates_dir + "coordinates_test.dat", overwrite=True
        )


class TestGridCoordinates:
    def test__from_yx_1d(self):
        coordinates = aa.GridCoordinates.from_yx_1d(y=[1.0, 1.0], x=[-1.0, 1.0])

        assert coordinates.in_list == [[(1.0, -1.0), (1.0, 1.0)]]

        coordinates = aa.GridCoordinates.from_yx_1d(
            y=[1.0, 1.0, 2.0, 4.0], x=[-1.0, 1.0, 3.0, 5.0]
        )

        assert coordinates.in_list == [
            [(1.0, -1.0), (1.0, 1.0), (2.0, 3.0), (4.0, 5.0)]
        ]

    def test__grid_from_deflection_grid(self):
        coordinates = aa.GridCoordinates(coordinates=[[(1.0, 1.0), (2.0, 2.0)]])

        coordinates = coordinates.grid_from_deflection_grid(
            deflection_grid=np.array([[1.0, 0.0], [1.0, 1.0]])
        )

        assert coordinates.in_list == [[(0.0, 1.0), (1.0, 1.0)]]

    def test__coordinates_from_grid_1d(self):

        coordinates = aa.GridCoordinates(coordinates=[[(1.0, 1.0), (2.0, 2.0)]])

        coordinates_from_1d = coordinates.coordinates_from_grid_1d(
            grid_1d=np.array([[1.0, 1.0], [2.0, 2.0]])
        )

        assert type(coordinates_from_1d) == grids.GridCoordinates
        assert coordinates_from_1d.in_list == [[(1.0, 1.0), (2.0, 2.0)]]

        coordinates = aa.GridCoordinates(
            coordinates=[[(1.0, 1.0), (2.0, 2.0)], [(3.0, 3.0)]]
        )

        coordinates_from_1d = coordinates.coordinates_from_grid_1d(
            grid_1d=np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]])
        )

        assert type(coordinates_from_1d) == grids.GridCoordinates
        assert coordinates_from_1d.in_list == [[(1.0, 1.0), (2.0, 2.0)], [(3.0, 3.0)]]

        coordinates_from_1d = coordinates.coordinates_from_grid_1d(
            grid_1d=np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]])
        )

        assert type(coordinates_from_1d) == grids.GridCoordinates
        assert coordinates_from_1d.in_list == [[(1.0, 1.0), (2.0, 2.0)], [(3.0, 3.0)]]

    def test__structure_from_result__maps_numpy_array_to__auto_array_or_grid(self):
        coordinates = aa.GridCoordinates(coordinates=[(1.0, -1.0), (1.0, 1.0)])

        result = coordinates.structure_from_result(result=np.array([1.0, 2.0]))

        assert isinstance(result, aa.Values)
        assert result.in_list == [[1.0, 2.0]]

        result = coordinates.structure_from_result(
            result=np.array([[1.0, 1.0], [2.0, 2.0]])
        )

        assert isinstance(result, aa.GridCoordinates)
        assert result.in_list == [[(1.0, 1.0), (2.0, 2.0)]]

    def test__structure_list_from_result_list__maps_list_to_auto_arrays_or_grids(self):
        coordinates = aa.GridCoordinates(coordinates=[(1.0, -1.0), (1.0, 1.0)])

        result = coordinates.structure_from_result(result=[np.array([1.0, 2.0])])

        assert isinstance(result[0], aa.Values)
        assert result[0].in_list == [[1.0, 2.0]]

        result = coordinates.structure_from_result(
            result=[np.array([[1.0, 1.0], [2.0, 2.0]])]
        )

        assert isinstance(result[0], aa.GridCoordinates)
        assert result[0].in_list == [[(1.0, 1.0), (2.0, 2.0)]]


class TestGridCoordinatesUniform:
    def test__from_sparse_uniform_upscale(self):

        grid_sparse_uniform = aa.GridCoordinatesUniform(
            coordinates=[[(1.0, 1.0), (1.0, 3.0)]], pixel_scales=2.0
        )

        grid_upscale = grids.GridCoordinatesUniform.from_grid_sparse_uniform_upscale(
            grid_sparse_uniform=grid_sparse_uniform, upscale_factor=2, pixel_scales=2.0
        )

        assert (
            grid_upscale
            == np.array(
                [
                    [1.5, 0.5],
                    [1.5, 1.5],
                    [0.5, 0.5],
                    [0.5, 1.5],
                    [1.5, 2.5],
                    [1.5, 3.5],
                    [0.5, 2.5],
                    [0.5, 3.5],
                ]
            )
        ).all()

        grid_sparse_uniform = aa.GridCoordinatesUniform(
            coordinates=[[(1.0, 1.0), (1.0, 3.0), (1.0, 5.0), (3.0, 3.0)]],
            pixel_scales=2.0,
        )

        grid_upscale = grids.GridCoordinatesUniform.from_grid_sparse_uniform_upscale(
            grid_sparse_uniform=grid_sparse_uniform, upscale_factor=4, pixel_scales=2.0
        )

        grid_upscale_util = aa.util.grid.grid_upscaled_1d_from(
            grid_1d=grid_sparse_uniform, upscale_factor=4, pixel_scales=(2.0, 2.0)
        )

        assert (grid_upscale == grid_upscale_util).all()

    def test__grid_from_deflection_grid(self):
        coordinates = aa.GridCoordinatesUniform(
            coordinates=[[(1.0, 1.0), (2.0, 2.0)]],
            pixel_scales=(1.0, 1.0),
            shape_2d=(3, 3),
        )

        coordinates = coordinates.grid_from_deflection_grid(
            deflection_grid=np.array([[1.0, 0.0], [1.0, 1.0]])
        )

        assert type(coordinates) == grids.GridCoordinatesUniform
        assert coordinates.in_list == [[(0.0, 1.0), (1.0, 1.0)]]
        assert coordinates.pixel_scales == (1.0, 1.0)
        assert coordinates.shape_2d == (3, 3)

    def test__coordinates_from_grid_1d(self):

        coordinates = aa.GridCoordinatesUniform(
            coordinates=[[(1.0, 1.0), (2.0, 2.0)]],
            pixel_scales=(1.0, 1.0),
            shape_2d=(3, 3),
        )

        coordinates_from_1d = coordinates.coordinates_from_grid_1d(
            grid_1d=np.array([[1.0, 1.0], [2.0, 2.0]])
        )

        assert type(coordinates_from_1d) == grids.GridCoordinatesUniform
        assert coordinates_from_1d.in_list == [[(1.0, 1.0), (2.0, 2.0)]]
        assert coordinates.pixel_scales == (1.0, 1.0)
        assert coordinates.shape_2d == (3, 3)

        coordinates = aa.GridCoordinatesUniform(
            coordinates=[[(1.0, 1.0), (2.0, 2.0)], [(3.0, 3.0)]],
            pixel_scales=(1.0, 3.0),
            shape_2d=(5, 5),
        )

        coordinates_from_1d = coordinates.coordinates_from_grid_1d(
            grid_1d=np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]])
        )

        assert type(coordinates_from_1d) == grids.GridCoordinatesUniform
        assert coordinates_from_1d.in_list == [[(1.0, 1.0), (2.0, 2.0)], [(3.0, 3.0)]]
        assert coordinates.pixel_scales == (1.0, 3.0)
        assert coordinates.shape_2d == (5, 5)



import logging

import numpy as np
import copy

from autoarray import exc
from autoarray.dataset import abstract_dataset, preprocess
from autoarray.mask import mask as msk
from autoarray.structures import arrays, grids, kernel
from autoarray.operators import convolver

logger = logging.getLogger(__name__)


class AbstractImaging(abstract_dataset.AbstractDataset):
    def __init__(self, image, noise_map, psf=None, positions=None, name=None):
        """A class containing the data, noise-map and point spread function of a 2D imaging dataset.

        Parameters
        ----------
        image : aa.Array
            The array of the image data, in units of electrons per second.
        noise_map : aa.Array
            An array describing the RMS standard deviation error in each pixel in units of electrons per second.
        psf : aa.Array
            An array describing the Point Spread Function kernel of the image.
        """

        super().__init__(
            data=image, noise_map=noise_map, positions=positions, name=name
        )

        self.psf = psf

    def __array_finalize__(self, obj):
        if isinstance(obj, Imaging):
            try:
                for key, value in obj.__dict__.items():
                    setattr(self, key, value)
            except AttributeError:
                logger.debug(
                    "Original object in Imaging.__array_finalize__ missing one or more attributes"
                )

    @property
    def shape_2d(self):
        return self.data.shape_2d

    @property
    def image(self):
        return self.data

    @property
    def pixel_scales(self):
        return self.data.pixel_scales

    def binned_from_bin_up_factor(self, bin_up_factor):

        imaging = copy.deepcopy(self)

        imaging.data = self.image.binned_from_bin_up_factor(
            bin_up_factor=bin_up_factor, method="mean"
        )
        imaging.psf = self.psf.rescaled_with_odd_dimensions_from_rescale_factor(
            rescale_factor=1.0 / bin_up_factor, renormalize=False
        )
        imaging.noise_map = (
            self.noise_map.binned_from_bin_up_factor(
                bin_up_factor=bin_up_factor, method="quadrature"
            )
            if self.noise_map is not None
            else None
        )

        return imaging

    def signal_to_noise_limited_from_signal_to_noise_limit(self, signal_to_noise_limit):

        imaging = copy.deepcopy(self)

        noise_map_limit = np.where(
            self.signal_to_noise_map > signal_to_noise_limit,
            np.abs(self.image) / signal_to_noise_limit,
            self.noise_map,
        )

        imaging.noise_map = arrays.Array.manual_mask(
            array=noise_map_limit,
            mask=self.image.mask,
            store_in_1d=self.noise_map.store_in_1d,
        )

        return imaging


class AbstractMaskedImaging(abstract_dataset.AbstractMaskedDataset):
    def __init__(
        self,
        imaging,
        mask,
        grid_class=grids.Grid,
        grid_inversion_class=grids.Grid,
        fractional_accuracy=0.9999,
        sub_steps=None,
        pixel_scales_interp=None,
        psf_shape_2d=None,
        inversion_pixel_limit=None,
        inversion_uses_border=True,
        renormalize_psf=True,
    ):
        """
        The lens dataset is the collection of data_type (image, noise-map, PSF), a mask, grid, convolver \
        and other utilities that are used for modeling and fitting an image of a strong lens.

        Whilst the image, noise-map, etc. are loaded in 2D, the lens dataset creates reduced 1D arrays of each \
        for lens calculations.

        Parameters
        ----------
        imaging: im.Imaging
            The imaging data_type all in 2D (the image, noise-map, PSF, etc.)
        mask: msk.Mask
            The 2D mask that is applied to the image.
        psf_shape_2d : (int, int)
            The shape of the PSF used for convolving model image generated using analytic light profiles. A smaller \
            shape will trim the PSF relative to the input image PSF, giving a faster analysis run-time.
        pixel_scales_interp : float
            If *True*, expensive to compute mass profile deflection angles will be computed on a sparse grid and \
            interpolated to the grid, sub and blurring grids.
        inversion_pixel_limit : int or None
            The maximum number of pixels that can be used by an inversion, with the limit placed primarily to speed \
            up run.
        """

        super().__init__(
            dataset=imaging,
            mask=mask,
            grid_class=grid_class,
            grid_inversion_class=grid_inversion_class,
            fractional_accuracy=fractional_accuracy,
            sub_steps=sub_steps,
            pixel_scales_interp=pixel_scales_interp,
            inversion_pixel_limit=inversion_pixel_limit,
            inversion_uses_border=inversion_uses_border,
        )

        self.image = arrays.Array.manual_mask(
            array=imaging.image.in_2d,
            mask=mask.mask_sub_1,
            store_in_1d=imaging.image.store_in_1d,
        )

        self.noise_map = arrays.Array.manual_mask(
            array=imaging.noise_map.in_2d,
            mask=mask.mask_sub_1,
            store_in_1d=imaging.noise_map.store_in_1d,
        )

        self.pixel_scales_interp = pixel_scales_interp

        ### PSF TRIMMING + CONVOLVER ###

        if imaging.psf is not None:

            if psf_shape_2d is None:
                self.psf_shape_2d = imaging.psf.shape_2d
            else:
                self.psf_shape_2d = psf_shape_2d

            self.psf = kernel.Kernel.manual_2d(
                array=imaging.psf.resized_from_new_shape(
                    new_shape=self.psf_shape_2d
                ).in_2d,
                renormalize=renormalize_psf,
            )

            self.convolver = convolver.Convolver(mask=mask, kernel=self.psf)

            if mask.pixel_scales is not None:

                self.blurring_grid = self.grid.blurring_grid_from_kernel_shape(
                    kernel_shape_2d=self.psf_shape_2d
                )

        else:

            self.psf = None

    @property
    def imaging(self):
        return self.dataset

    @property
    def data(self):
        return self.image

    def signal_to_noise_map(self):
        return self.image / self.noise_map

    def modify_image_and_noise_map(self, image, noise_map):

        masked_imaging = copy.deepcopy(self)

        masked_imaging.image = image
        masked_imaging.noise_map = noise_map

        return masked_imaging


class AbstractSimulatorImaging:
    def __init__(
        self,
        exposure_time_map=None,
        background_sky_map=None,
        psf=None,
        renormalize_psf=True,
        read_noise=None,
        add_noise=True,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
    ):
        """A class representing a Imaging observation, using the shape of the image, the pixel scale,
        psf, exposure time, etc.

        Parameters
        ----------
        shape_2d : (int, int)
            The shape of the observation. Note that we do not simulator a full Imaging frame (e.g. 2000 x 2000 pixels for \
            Hubble imaging), but instead just a cut-out around the strong lens.
        pixel_scales : float
            The size of each pixel in arc seconds.
        psf : PSF
            An arrays describing the PSF kernel of the image.
        exposure_time_map : float
            The exposure time of an observation using this data_type.
        background_sky_map : float
            The level of the background sky of an observationg using this data_type.
        """

        self.exposure_time_map = exposure_time_map

        if psf is not None and renormalize_psf:
            psf = psf.renormalized

        self.psf = psf
        self.background_sky_map = background_sky_map
        self.read_noise = read_noise
        self.add_noise = add_noise
        self.noise_if_add_noise_false = noise_if_add_noise_false
        self.noise_seed = noise_seed


class Imaging(AbstractImaging):
    @classmethod
    def from_fits(
        cls,
        image_path,
        pixel_scales=None,
        image_hdu=0,
        noise_map_path=None,
        noise_map_hdu=0,
        psf_path=None,
        psf_hdu=0,
        positions_path=None,
        name=None,
    ):
        """Factory for loading the imaging data_type from .fits files, as well as computing properties like the noise-map,
        exposure-time map, etc. from the imaging-data.

        This factory also includes a number of routines for converting the imaging-data from unit_label not supported by PyAutoLens \
        (e.g. adus, electrons) to electrons per second.

        Parameters
        ----------
        renormalize_psf
        noise_map_non_constant
        name
        image_path : str
            The path to the image .fits file containing the image (e.g. '/path/to/image.fits')
        pixel_scales : float
            The size of each pixel in arc seconds.
        image_hdu : int
            The hdu the image is contained in the .fits file specified by *image_path*.
        psf_path : str
            The path to the psf .fits file containing the psf (e.g. '/path/to/psf.fits')
        psf_hdu : int
            The hdu the psf is contained in the .fits file specified by *psf_path*.
        noise_map_path : str
            The path to the noise_map .fits file containing the noise_map (e.g. '/path/to/noise_map.fits')
        noise_map_hdu : int
            The hdu the noise_map is contained in the .fits file specified by *noise_map_path*.
        """

        image = arrays.Array.from_fits(
            file_path=image_path, hdu=image_hdu, pixel_scales=pixel_scales
        )

        noise_map = arrays.Array.from_fits(
            file_path=noise_map_path, hdu=noise_map_hdu, pixel_scales=pixel_scales
        )

        if psf_path is not None:

            psf = kernel.Kernel.from_fits(
                file_path=psf_path,
                hdu=psf_hdu,
                pixel_scales=pixel_scales,
                renormalize=True,
            )

        else:

            psf = None

        if positions_path is not None:

            positions = grids.GridCoordinates.from_file(file_path=positions_path)

        else:

            positions = None

        return Imaging(
            image=image, noise_map=noise_map, psf=psf, positions=positions, name=name
        )

    def output_to_fits(
        self, image_path, psf_path=None, noise_map_path=None, overwrite=False
    ):
        self.image.output_to_fits(file_path=image_path, overwrite=overwrite)

        if self.psf is not None and psf_path is not None:
            self.psf.output_to_fits(file_path=psf_path, overwrite=overwrite)

        if self.noise_map is not None and noise_map_path is not None:
            self.noise_map.output_to_fits(file_path=noise_map_path, overwrite=overwrite)


class MaskedImaging(AbstractMaskedImaging):

    pass


class SimulatorImaging(AbstractSimulatorImaging):
    def from_image(self, image, name=None):
        """
        Create a realistic simulated image by applying effects to a plain simulated image.

        Parameters
        ----------
        noise_if_add_noise_false
        background_level
        exposure_time_
        name
        image : ndarray
            The image before simulating (e.g. the lens and source galaxies before optics blurring and Imaging read-out).
        exposure_time_map : ndarray
            An array representing the effective exposure time of each pixel.
        psf: PSF
            An array describing the PSF the simulated image is blurred with.
        background_sky_map : ndarray
            The value of background sky in every image pixel (electrons per second).
        add_noise: Bool
            If True poisson noise_maps is simulated and added to the image, based on the total counts in each image
            pixel
        noise_seed: int
            A seed for random noise_maps generation
        """

        if self.background_sky_map is not None:
            background_sky_map = self.background_sky_map
        else:
            background_sky_map = arrays.Array.zeros(
                shape_2d=image.shape_2d, pixel_scales=image.pixel_scales
            )

        if self.psf is not None:
            psf = self.psf
        else:
            psf = kernel.Kernel.no_blur(pixel_scales=image.pixel_scales)

        image = psf.convolved_array_from_array(array=image)

        image = image + background_sky_map

        image = image.trimmed_from_kernel_shape(kernel_shape_2d=psf.shape_2d)
        exposure_time_map = self.exposure_time_map.trimmed_from_kernel_shape(
            kernel_shape_2d=psf.shape_2d
        )
        background_sky_map = background_sky_map.trimmed_from_kernel_shape(
            kernel_shape_2d=psf.shape_2d
        )

        if self.add_noise is True:
            image = preprocess.data_with_poisson_noise_added(
                data=image, exposure_time_map=exposure_time_map, seed=self.noise_seed
            )

            noise_map = preprocess.noise_map_from_data_and_exposure_time_map(
                data=image, exposure_time_map=exposure_time_map
            )

        else:
            noise_map = arrays.Array.full(
                fill_value=self.noise_if_add_noise_false,
                shape_2d=image.shape_2d,
                pixel_scales=image.pixel_scales,
            )

        if np.isnan(noise_map).any():
            raise exc.DataException(
                "The noise-map has NaN values in it. This suggests your exposure time and / or"
                "background sky levels are too low, creating signal counts at or close to 0.0."
            )

        image = image - background_sky_map

        mask = msk.Mask.unmasked(
            shape_2d=image.shape_2d, pixel_scales=image.pixel_scales
        )

        image = arrays.Array.manual_mask(array=image, mask=mask)

        return Imaging(image=image, psf=self.psf, noise_map=noise_map, name=name)



import numpy as np


def residual_map_with_mask_from(*, data, mask, model_data):
    """Compute the residual-map between a masked dataset and model data, where:

    Residuals = (Data - Model_Data).
    
    The residual-map values in masked pixels are returned as zero.

    Parameters
    -----------
    data : np.ndarray
        The data that is fitted.
    mask : np.ndarray
        The mask applied to the dataset, where *False* entries are included in the calculation.
    model_data : np.ndarray
        The model data used to fit the data.
    """
    return np.subtract(
        data, model_data, out=np.zeros_like(data), where=np.asarray(mask) == 0
    )


def normalized_residual_map_with_mask_from(*, residual_map, noise_map, mask):
    """Compute the normalized residual-map between a masked dataset and model data, where:

    Normalized_Residual = (Data - Model_Data) / Noise

    The normalized residual-map values in masked pixels are returned as zero.

    Parameters
    -----------
    residual_map : np.ndarray
        The residual-map of the model-simulator fit to the dataset.
    noise_map : np.ndarray
        The noise-map of the dataset.
    mask : np.ndarray
        The mask applied to the residual-map, where *False* entries are included in the calculation.
    """
    return np.divide(
        residual_map,
        noise_map,
        out=np.zeros_like(residual_map),
        where=np.asarray(mask) == 0,
    )


def chi_squared_map_with_mask_from(*, residual_map, noise_map, mask):
    """Computes the chi-squared-map between a masked residual-map and noise-map, where:

    Chi_Squared = ((Residuals) / (Noise)) ** 2.0 = ((Data - Model)**2.0)/(Variances)

    The chi-squared-map values in masked pixels are returned as zero.

    Parameters
    -----------
    residual_map : np.ndarray
        The residual-map of the model-simulator fit to the dataset.
    noise_map : np.ndarray
        The noise-map of the dataset.
    mask : np.ndarray
        The mask applied to the residual-map, where *False* entries are included in the calculation.
    """
    return np.square(
        np.divide(
            residual_map,
            noise_map,
            out=np.zeros_like(residual_map),
            where=np.asarray(mask) == 0,
        )
    )


def chi_squared_with_mask_from(*, chi_squared_map, mask):
    """Compute the chi-squared terms of each model data's fit to a masked dataset, by summing the masked
    chi-squared-map of the fit.

    The chi-squared values in masked pixels are omitted from the calculation.

    Parameters
    ----------
    chi_squared_map : np.ndarray
        The chi-squared-map of values of the model-simulator fit to the dataset.
    mask : np.ndarray
        The mask applied to the chi-squared-map, where *False* entries are included in the calculation.
    """
    return np.sum(chi_squared_map[np.asarray(mask) == 0])


def noise_normalization_with_mask_from(*, noise_map, mask):
    """Compute the noise-map normalization terms of masked noise-map, summing the noise_map value in every pixel as:

    [Noise_Term] = sum(log(2*pi*[Noise]**2.0))

    The noise-map values in masked pixels are omitted from the calculation.

    Parameters
    ----------
    noise_map : np.ndarray
        The masked noise-map of the dataset.
    mask : np.ndarray
        The mask applied to the noise-map, where *False* entries are included in the calculation.
    """
    return np.sum(np.log(2 * np.pi * noise_map[np.asarray(mask) == 0] ** 2.0))


def residual_map_from(*, data, model_data):
    """Compute the residual-map between a masked dataset and model data, where:

    Residuals = (Data - Model_Data).

    Parameters
    -----------
    data : np.ndarray
        The data that is fitted.
    mask : np.ndarray
        The mask applied to the dataset, where *False* entries are included in the calculation.
    model_data : np.ndarray
        The model data used to fit the data.
    """
    return np.subtract(data, model_data, out=np.zeros_like(data))


def normalized_residual_map_from(*, residual_map, noise_map):
    """Compute the normalized residual-map between a masked dataset and model data, where:

    Normalized_Residual = (Data - Model_Data) / Noise

    Parameters
    -----------
    residual_map : np.ndarray
        The residual-map of the model-simulator fit to the dataset.
    noise_map : np.ndarray
        The noise-map of the dataset.
    mask : np.ndarray
        The mask applied to the residual-map, where *False* entries are included in the calculation.
    """
    return np.divide(residual_map, noise_map, out=np.zeros_like(residual_map))


def chi_squared_map_from(*, residual_map, noise_map):
    """Compute the chi-squared-map between a residual-map and noise-map, where:

    Chi_Squared = ((Residuals) / (Noise)) ** 2.0 = ((Data - Model)**2.0)/(Variances)

    Parameters
    -----------
    residual_map : np.ndarray
        The residual-map of the model-simulator fit to the dataset.
    noise_map : np.ndarray
        The noise-map of the dataset.
    """
    return np.square(
        np.divide(residual_map, noise_map, out=np.zeros_like(residual_map))
    )


def chi_squared_from(*, chi_squared_map):
    """Compute the chi-squared terms of a model data's fit to an dataset, by summing the chi-squared-map.

    Parameters
    ----------
    chi_squared_map : np.ndarray
        The chi-squared-map of values of the model-simulator fit to the dataset.
    """
    return np.sum(chi_squared_map)


def noise_normalization_from(*, noise_map):
    """Compute the noise-map normalization term of the noise-map, summing the noise_map value in every pixel as:

    [Noise_Term] = sum(log(2*pi*[Noise]**2.0))

    Parameters
    ----------
    noise_map : np.ndarray
        The masked noise-map of the dataset.
    """
    return np.sum(np.log(2 * np.pi * noise_map ** 2.0))


def log_likelihood_from(*, chi_squared, noise_normalization):
    """Compute the log likelihood of each model data point's fit to the dataset, where:

    Log Likelihood = -0.5*[Chi_Squared_Term + Noise_Term] (see functions above for these definitions)

    Parameters
    ----------
    chi_squared : float
        The chi-squared term for the model-simulator fit to the dataset.
    noise_normalization : float
        The normalization noise_map-term for the dataset's noise-map.
    """
    return -0.5 * (chi_squared + noise_normalization)


def log_likelihood_with_regularization_from(
    *, chi_squared, regularization_term, noise_normalization
):
    """Compute the log likelihood of an inversion's fit to the dataset, including a regularization term which \
    comes from an inversion:

    Log Likelihood = -0.5*[Chi_Squared_Term + Regularization_Term + Noise_Term] (see functions above for these definitions)

    Parameters
    ----------
    chi_squared : float
        The chi-squared term of the inversion's fit to the dataset.
    regularization_term : float
        The regularization term of the inversion, which is the sum of the difference between reconstructed \
        flux of every pixel multiplied by the regularization coefficient.
    noise_normalization : float
        The normalization noise_map-term for the dataset's noise-map.
    """
    return -0.5 * (chi_squared + regularization_term + noise_normalization)


def log_evidence_from(
    *,
    chi_squared,
    regularization_term,
    log_curvature_regularization_term,
    log_regularization_term,
    noise_normalization,
):
    """Compute the log evidence of an inversion's fit to a dataset, where the log evidence includes a number of terms which
    quantify the complexity of an inversion's reconstruction (see the *inversion* module):

    Log Evidence = -0.5*[Chi_Squared_Term + Regularization_Term + Log(Covariance_Regularization_Term) -
                           Log(Regularization_Matrix_Term) + Noise_Term]

    Parameters
    ----------
    chi_squared : float
        The chi-squared term of the inversion's fit to the dataset.
    regularization_term : float
        The regularization term of the inversion, which is the sum of the difference between reconstructed \
        flux of every pixel multiplied by the regularization coefficient.
    log_curvature_regularization_term : float
        The log of the determinant of the sum of the curvature and regularization matrices.
    log_regularization_term : float
        The log of the determinant o the regularization matrix.
    noise_normalization : float
        The normalization noise_map-term for the dataset's noise-map.
    """
    return -0.5 * (
        chi_squared
        + regularization_term
        + log_curvature_regularization_term
        - log_regularization_term
        + noise_normalization
    )



import autoarray as aa
import autoarray.plot as aplt

array = aa.Array.ones(shape_2d=(31, 31), pixel_scales=(1.0, 1.0), sub_size=2)
array[0] = 3.0

mask = aa.Mask.circular(
    shape_2d=array.shape_2d,
    pixel_scales=array.pixel_scales,
    radius=5.0,
    centre=(2.0, 2.0),
)

aplt.Array(array=array, mask=mask, include_border=True)
